from __future__ import annotations

import datetime
import re
import string
import unicodedata

import pyarrow as pa
import pytest

import daft
import daft.exceptions
from daft import DataType, Series


@pytest.mark.parametrize(
    ["funcname", "data"],
    [
        ("endswith", ["x_foo", "y_foo", "z_bar"]),
        ("startswith", ["foo_x", "foo_y", "bar_z"]),
        ("contains", ["x_foo_x", "y_foo_y", "z_bar_z"]),
    ],
)
def test_series_utf8_compare(funcname, data) -> None:
    s = Series.from_arrow(pa.array(data))
    pattern = Series.from_arrow(pa.array(["foo", "bar", "bar"]))
    result = getattr(s.str, funcname)(pattern)
    assert result.to_pylist() == [True, False, True]


@pytest.mark.parametrize(
    ["funcname", "data"],
    [
        ("endswith", ["x_foo", "y_foo", "z_bar"]),
        ("startswith", ["foo_x", "foo_y", "bar_z"]),
        ("contains", ["x_foo_x", "y_foo_y", "z_bar_z"]),
    ],
)
def test_series_utf8_compare_pattern_broadcast(funcname, data) -> None:
    s = Series.from_arrow(pa.array(data))
    pattern = Series.from_arrow(pa.array(["foo"]))
    result = getattr(s.str, funcname)(pattern)
    assert result.to_pylist() == [True, True, False]


@pytest.mark.parametrize(
    ["funcname", "data"],
    [
        ("endswith", ["x_foo", "y_foo", "z_bar"]),
        ("startswith", ["foo_x", "foo_y", "bar_z"]),
        ("contains", ["x_foo_x", "y_foo_y", "z_bar_z"]),
    ],
)
def test_series_utf8_compare_pattern_broadcast_null(funcname, data) -> None:
    s = Series.from_arrow(pa.array(["x_foo", "y_foo", "z_bar"]))
    pattern = Series.from_arrow(pa.array([None], type=pa.string()))
    result = getattr(s.str, funcname)(pattern)
    assert result.to_pylist() == [None, None, None]


@pytest.mark.parametrize(
    ["funcname", "data"], [("endswith", ["x_foo"]), ("startswith", ["foo_x"]), ("contains", ["x_foo_x"])]
)
def test_series_utf8_compare_data_broadcast(funcname, data) -> None:
    s = Series.from_arrow(pa.array(data))
    pattern = Series.from_arrow(pa.array(["foo", "bar", "baz"]))
    result = getattr(s.str, funcname)(pattern)
    assert result.to_pylist() == [True, False, False]


@pytest.mark.parametrize("funcname", ["endswith", "startswith", "contains"])
def test_series_utf8_compare_data_broadcast_null(funcname) -> None:
    s = Series.from_arrow(pa.array([None], type=pa.string()))
    pattern = Series.from_arrow(pa.array(["foo", "bar", "baz"]))
    result = getattr(s.str, funcname)(pattern)
    assert result.to_pylist() == [None, None, None]


@pytest.mark.parametrize("funcname", ["endswith", "startswith", "contains"])
def test_series_utf8_compare_nulls(funcname) -> None:
    s = Series.from_arrow(pa.array([None, None, "z_bar"]))
    pattern = Series.from_arrow(pa.array([None, "bar", None]))
    result = getattr(s.str, funcname)(pattern)
    assert result.to_pylist() == [None, None, None]


@pytest.mark.parametrize("funcname", ["endswith", "startswith", "contains"])
def test_series_utf8_compare_empty(funcname) -> None:
    s = Series.from_arrow(pa.array([], type=pa.string()))
    pattern = Series.from_arrow(pa.array([], type=pa.string()))
    result = getattr(s.str, funcname)(pattern)
    assert result.to_pylist() == []


@pytest.mark.parametrize(
    "bad_series",
    [
        # Wrong number of elements, not broadcastable
        Series.from_arrow(pa.array([], type=pa.string())),
        Series.from_arrow(pa.array(["foo", "bar"], type=pa.string())),
        # Bad input type
        object(),
    ],
)
@pytest.mark.parametrize("funcname", ["endswith", "startswith", "contains"])
def test_series_utf8_compare_invalid_inputs(funcname, bad_series) -> None:
    s = Series.from_arrow(pa.array(["x_foo", "y_foo", "z_bar"]))
    with pytest.raises(ValueError):
        getattr(s.str, funcname)(bad_series)


@pytest.mark.parametrize(
    ["data", "patterns", "expected", "regex"],
    [
        # Single-character pattern.
        (["a,b,c", "d,e", "f", "g,h"], [","], [["a", "b", "c"], ["d", "e"], ["f"], ["g", "h"]], False),
        # Multi-character pattern.
        (["abbcbbd", "bb", "bbe", "fbb"], ["bb"], [["a", "c", "d"], ["", ""], ["", "e"], ["f", ""]], False),
        # Empty pattern (character-splitting).
        (["foo", "bar"], [""], [["", "f", "o", "o", ""], ["", "b", "a", "r", ""]], False),
        # Single-character pattern (regex).
        (["a,b,c", "d,e", "f", "g,h"], [r","], [["a", "b", "c"], ["d", "e"], ["f"], ["g", "h"]], True),
        # Multi-character pattern (regex).
        (["abbcbbd", "bb", "bbe", "fbb"], [r"b+"], [["a", "c", "d"], ["", ""], ["", "e"], ["f", ""]], True),
        # Empty pattern (regex).
        (["foo", "bar"], [r""], [["", "f", "o", "o", ""], ["", "b", "a", "r", ""]], True),
    ],
)
def test_series_utf8_split_broadcast_pattern(data, patterns, expected, regex) -> None:
    s = Series.from_arrow(pa.array(data))
    patterns = Series.from_arrow(pa.array(patterns))
    result = s.str.split(patterns, regex=regex)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "patterns", "expected", "regex"],
    [
        (["a,b,c", "a:b:c", "a;b;c", "a.b.c"], [",", ":", ";", "."], [["a", "b", "c"]] * 4, False),
        (
            ["aabbccdd"] * 4,
            ["aa", "bb", "cc", "dd"],
            [["", "bbccdd"], ["aa", "ccdd"], ["aabb", "dd"], ["aabbcc", ""]],
            False,
        ),
        (["a,b,c", "a:b:c", "a;b;c", "a.b.c"], [r",", r":", r";", r"\."], [["a", "b", "c"]] * 4, True),
        (
            ["aabbccdd"] * 4,
            [r"a+", r"b+", r"c+", r"d+"],
            [["", "bbccdd"], ["aa", "ccdd"], ["aabb", "dd"], ["aabbcc", ""]],
            True,
        ),
    ],
)
def test_series_utf8_split_multi_pattern(data, patterns, expected, regex) -> None:
    s = Series.from_arrow(pa.array(data))
    patterns = Series.from_arrow(pa.array(patterns))
    result = s.str.split(patterns, regex=regex)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "patterns", "expected", "regex"],
    [
        (
            ["aabbccdd"],
            ["aa", "bb", "cc", "dd"],
            [["", "bbccdd"], ["aa", "ccdd"], ["aabb", "dd"], ["aabbcc", ""]],
            False,
        ),
        (
            ["aabbccdd"],
            [r"a+", r"b+", r"c+", r"d+"],
            [["", "bbccdd"], ["aa", "ccdd"], ["aabb", "dd"], ["aabbcc", ""]],
            True,
        ),
    ],
)
def test_series_utf8_split_broadcast_arr(data, patterns, expected, regex) -> None:
    s = Series.from_arrow(pa.array(data))
    patterns = Series.from_arrow(pa.array(patterns))
    result = s.str.split(patterns, regex=regex)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "patterns", "expected"],
    [
        # Mixed-in nulls.
        (["a,b,c", None, "a;b;c", "a/b/c"], [",", ":", None, "/"], [["a", "b", "c"], None, None, ["a", "b", "c"]]),
        # All null data.
        ([None] * 4, [","] * 4, [None] * 4),
        # All null patterns.
        (["foo"] * 4, [None] * 4, [None] * 4),
        # Broadcasted null data.
        ([None], [","] * 4, [None] * 4),
        # Broadcasted null pattern.
        (["foo"] * 4, [None], [None] * 4),
    ],
)
@pytest.mark.parametrize("regex", [False, True])
def test_series_utf8_split_nulls(data, patterns, expected, regex) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(patterns, type=pa.string()))
    result = s.str.split(patterns, regex=regex)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "patterns"],
    [
        # Empty patterns.
        ([["foo"] * 4, []]),
    ],
)
@pytest.mark.parametrize("regex", [False, True])
def test_series_utf8_split_empty_arrs(data, patterns, regex) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(patterns, type=pa.string()))
    with pytest.raises(ValueError):
        s.str.split(patterns, regex=regex)


def test_series_utf8_split_empty_input() -> None:
    s = Series.from_arrow(pa.array([], type=pa.string()))
    patterns = Series.from_arrow(pa.array([","], type=pa.string()))
    assert len(s.str.split(patterns)) == 0


@pytest.mark.parametrize(
    "patterns",
    [
        # Wrong number of elements, not broadcastable
        Series.from_arrow(pa.array([",", "."], type=pa.string())),
        # Bad input type
        object(),
    ],
)
@pytest.mark.parametrize("regex", [False, True])
def test_series_utf8_split_invalid_inputs(patterns, regex) -> None:
    s = Series.from_arrow(pa.array(["a,b,c", "d, e", "f"]))
    with pytest.raises(ValueError):
        s.str.split(patterns, regex=regex)


def test_series_utf8_length() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    result = s.str.length()
    assert result.to_pylist() == [3, 6, 4]


def test_series_utf8_length_with_nulls() -> None:
    s = Series.from_arrow(pa.array(["foo", None, "barbaz", "quux"]))
    result = s.str.length()
    assert result.to_pylist() == [3, None, 6, 4]


def test_series_utf8_length_empty() -> None:
    s = Series.from_arrow(pa.array([], type=pa.string()))
    result = s.str.length()
    assert result.to_pylist() == []


def test_series_utf8_length_all_null() -> None:
    s = Series.from_arrow(pa.array([None, None, None]))
    result = s.str.length()
    assert result.to_pylist() == [None, None, None]


def test_series_utf8_length_unicode() -> None:
    s = Series.from_arrow(pa.array(["😉test", "hey̆"]))
    result = s.str.length()
    assert result.to_pylist() == [5, 4]


def test_series_utf8_length_bytes() -> None:
    s = Series.from_arrow(pa.array(["😉test", "hey̆", "baz", None, ""]))
    result = s.str.length_bytes()
    assert result.to_pylist() == [8, 5, 3, None, 0]


@pytest.mark.parametrize(
    ["data", "expected"],
    [
        (["Foo", "BarBaz", "QUUX"], ["foo", "barbaz", "quux"]),
        # With at least one null
        (["Foo", None, "BarBaz", "QUUX"], ["foo", None, "barbaz", "quux"]),
        # With all nulls
        ([None] * 4, [None] * 4),
        # With at least one numeric strings
        (["Foo", "BarBaz", "QUUX", "2"], ["foo", "barbaz", "quux", "2"]),
        # With all numeric strings
        (["1", "2", "3"], ["1", "2", "3"]),
    ],
)
def test_series_utf8_lower(data, expected) -> None:
    s = Series.from_arrow(pa.array(data))
    result = s.str.lower()
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "expected"],
    [
        (["Foo", "BarBaz", "quux"], ["FOO", "BARBAZ", "QUUX"]),
        # With at least one null
        (["Foo", None, "BarBaz", "quux"], ["FOO", None, "BARBAZ", "QUUX"]),
        # With all nulls
        ([None] * 4, [None] * 4),
        # With at least one numeric strings
        (["Foo", "BarBaz", "quux", "2"], ["FOO", "BARBAZ", "QUUX", "2"]),
        # With all numeric strings
        (["1", "2", "3"], ["1", "2", "3"]),
    ],
)
def test_series_utf8_upper(data, expected) -> None:
    s = Series.from_arrow(pa.array(data))
    result = s.str.upper()
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "expected"],
    [
        (["\ta\t", "\nb\n", "\vc\t", "\td ", "e"], ["a\t", "b\n", "c\t", "d ", "e"]),
        # With at least one null
        (["\ta\t", None, "\vc\t", "\td ", "e"], ["a\t", None, "c\t", "d ", "e"]),
        # With all nulls
        ([None] * 4, [None] * 4),
    ],
)
def test_series_utf8_lstrip(data, expected) -> None:
    s = Series.from_arrow(pa.array(data))
    result = s.str.lstrip()
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "expected"],
    [
        (["\ta\t", "\nb\n", "\vc\t", "\td ", "e"], ["\ta", "\nb", "\vc", "\td", "e"]),
        # With at least one null
        (["\ta\t", None, "\vc\t", "\td ", "e"], ["\ta", None, "\vc", "\td", "e"]),
        # With all nulls
        ([None] * 4, [None] * 4),
    ],
)
def test_series_utf8_rstrip(data, expected) -> None:
    s = Series.from_arrow(pa.array(data))
    result = s.str.rstrip()
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "expected"],
    [
        (["abc", "def", "ghi"], ["cba", "fed", "ihg"]),
        # With at least one null
        (["abc", None, "def", "ghi"], ["cba", None, "fed", "ihg"]),
        # With all nulls
        ([None] * 4, [None] * 4),
        # With emojis
        (["😃😌😝", "abc😃😄😅"], ["😝😌😃", "😅😄😃cba"]),
        # With non-latin alphabet
        (["こんにちは", "こんにちはa", "こんにちはa😄😃"], ["はちにんこ", "aはちにんこ", "😃😄aはちにんこ"]),
    ],
)
def test_series_utf8_reverse(data, expected) -> None:
    s = Series.from_arrow(pa.array(data))
    result = s.str.reverse()
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "expected"],
    [
        (["Foo", "BarBaz", "quux"], ["Foo", "Barbaz", "Quux"]),
        # With at least one null
        (["Foo", None, "BarBaz", "quux"], ["Foo", None, "Barbaz", "Quux"]),
        # With all nulls
        ([None] * 4, [None] * 4),
        # With at least one numeric strings
        (["Foo", "BarBaz", "quux", "2"], ["Foo", "Barbaz", "Quux", "2"]),
        # With all numeric strings
        (["1", "2", "3"], ["1", "2", "3"]),
        # With empty string
        (["", "Foo", "BarBaz", "quux"], ["", "Foo", "Barbaz", "Quux"]),
        # With emojis
        (["😃😌😝", "abc😃😄😅"], ["😃😌😝", "Abc😃😄😅"]),
    ],
)
def test_series_utf8_capitalize(data, expected) -> None:
    s = Series.from_arrow(pa.array(data))
    result = s.str.capitalize()
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "pattern", "expected"],
    [
        # No broadcast
        (["foo", "barbaz", "quux"], ["foo", "bar", "baz"], [True, True, False]),
        # Broadcast pattern
        (["foo", "barbaz", "quux"], ["foo"], [True, False, False]),
        # Broadcast data
        (["foo"], ["foo", "bar", "baz"], [True, False, False]),
        # Broadcast null data
        ([None], ["foo", "bar", "baz"], [None, None, None]),
        # Broadcast null pattern
        (["foo", "barbaz", "quux"], [None], [None, None, None]),
        # Mixed-in nulls
        (["foo", None, "barbaz", "quux"], ["foo", "bar", "boo", None], [True, None, False, None]),
    ],
)
def test_series_utf8_match(data, pattern, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pattern, type=pa.string()))
    result = s.str.match(patterns)
    assert result.to_pylist() == expected


def test_series_utf8_match_bad_pattern() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pattern = Series.from_arrow(pa.array(["["]))
    with pytest.raises(ValueError):
        s.str.match(pattern)


@pytest.mark.parametrize(
    ["data", "nchars", "expected"],
    [
        # No Broadcast
        (["foo", "barbaz", "quux"], [0, 1, 2], ["", "b", "qu"]),
        # Broadcast nchars
        (["foo", "barbaz", "quux"], [1], ["f", "b", "q"]),
        # Broadcast data
        (["foo"], [0, 1, 2], ["", "f", "fo"]),
        # Broadcast null data
        ([None], [0, 1, 2], [None, None, None]),
        # Broadcast null nchars
        (["foo", "barbaz", "quux"], [None], [None, None, None]),
        # All Empty
        ([[], [], []]),
        # Mixed-in nulls
        (["foo", None, "barbaz", "quux"], [0, 1, 1, None], ["", None, "b", None]),
        # All null data.
        ([None] * 4, [1] * 4, [None] * 4),
        # All null nchars
        (["foo"] * 4, [None] * 4, [None] * 4),
        # Broadcasted null data.
        ([None], [1] * 4, [None] * 4),
        # Broadcasted null nchars
        (["foo"] * 4, [None], [None] * 4),
        # with emojis
        (["😃😌😝", "abc😃😄😅"], [1, 6], ["😃", "abc😃😄😅"]),
    ],
)
def test_series_utf8_left(data, nchars, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    nchars = Series.from_arrow(pa.array(nchars, type=pa.uint32()))
    result = s.str.left(nchars)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "nchars"],
    [
        # empty nchars
        ([["foo"] * 4, []]),
    ],
)
def test_series_utf8_left_empty_arrs(data, nchars) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    nchars = Series.from_arrow(pa.array(nchars, type=pa.uint32()))
    with pytest.raises(daft.exceptions.DaftCoreException):
        s.str.left(nchars)


def test_series_utf8_left_empty_inputs() -> None:
    s = Series.from_arrow(pa.array([], type=pa.string()))
    nchars = Series.from_arrow(pa.array([5], type=pa.uint32()))
    assert len(s.str.left(nchars)) == 0


def test_series_utf8_left_mismatch_len() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    nchars = Series.from_arrow(pa.array([1, 2], type=pa.uint32()))
    with pytest.raises(daft.exceptions.DaftCoreException):
        s.str.left(nchars)


def test_series_utf8_left_bad_nchars_dtype() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    nchars = Series.from_arrow(pa.array(["1", "2", "3"]))
    with pytest.raises(daft.exceptions.DaftTypeError):
        s.str.left(nchars)


def test_series_utf8_left_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    nchars = Series.from_arrow(pa.array([1, 2, 3]))
    with pytest.raises(daft.exceptions.DaftTypeError):
        s.str.left(nchars)


@pytest.mark.parametrize(
    ["data", "nchars", "expected"],
    [
        (["foo", "barbaz", "quux"], [0, 1, 2], ["", "z", "ux"]),
        (["foo", "barbaz", "quux"], [1], ["o", "z", "x"]),
        (["foo"], [0, 1, 2], ["", "o", "oo"]),
        ([None], [0, 1, 2], [None, None, None]),
        (["foo", "barbaz", "quux"], [None], [None, None, None]),
        ([[], [], []]),
        (["foo", None, "barbaz", "quux"], [0, 1, 1, None], ["", None, "z", None]),
        ([None] * 4, [1] * 4, [None] * 4),
        (["foo"] * 4, [None] * 4, [None] * 4),
        ([None], [1] * 4, [None] * 4),
        (["foo"] * 4, [None], [None] * 4),
        (["😃😌😝", "abc😃😄😅"], [1, 6], ["😝", "abc😃😄😅"]),
    ],
)
def test_series_utf8_right(data, nchars, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    nchars = Series.from_arrow(pa.array(nchars, type=pa.uint32()))
    result = s.str.right(nchars)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "nchars"],
    [
        # empty nchars
        ([["foo"] * 4, []]),
    ],
)
def test_series_utf8_right_empty_arrs(data, nchars) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    nchars = Series.from_arrow(pa.array(nchars, type=pa.uint32()))
    with pytest.raises(ValueError):
        s.str.right(nchars)


def test_series_utf8_right_mismatch_len() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    nchars = Series.from_arrow(pa.array([1, 2], type=pa.uint32()))
    with pytest.raises(ValueError):
        s.str.right(nchars)


def test_series_utf8_right_bad_nchars_dtype() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    nchars = Series.from_arrow(pa.array(["1", "2", "3"]))
    with pytest.raises(ValueError):
        s.str.right(nchars)


def test_series_utf8_right_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    nchars = Series.from_arrow(pa.array([1, 2, 3]))
    with pytest.raises(ValueError):
        s.str.right(nchars)


@pytest.mark.parametrize(
    ["data", "pattern", "expected"],
    [
        # No broadcast
        (["123", "456", "789"], [r"\d+", r"\d+", r"\d"], ["123", "456", "7"]),
        # Broadcast pattern
        (["123", "456", "789"], [r"\d+"], ["123", "456", "789"]),
        # Broadcast data
        (["123"], [r"\d+", r"\d{2}", r"\d"], ["123", "12", "1"]),
        # Broadcast null data
        ([None], [r"\d+", r"\d{2}", r"\d"], [None, None, None]),
        # Broadcast null pattern
        (["123", "456", "789"], [None], [None, None, None]),
        # Mixed in nulls
        (["123", None, "789"], [None, r"\d+", r"\d"], [None, None, "7"]),
    ],
)
def test_series_utf8_extract(data, pattern, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pattern, type=pa.string()))
    result = s.str.extract(patterns)
    assert result.to_pylist() == expected


@pytest.mark.parametrize("data", [["abc abc"]])
@pytest.mark.parametrize("pattern", [["(a)bc"]])
@pytest.mark.parametrize(
    "index, expected",
    [
        (0, ["abc"]),
        (1, ["a"]),
        (2, [None]),
    ],
)
def test_series_utf8_extract_index(data, pattern, index, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pattern, type=pa.string()))
    result = s.str.extract(patterns, index)
    assert result.to_pylist() == expected


def test_series_utf8_extract_bad_pattern() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pattern = Series.from_arrow(pa.array(["["]))
    with pytest.raises(ValueError):
        s.str.extract(pattern)


@pytest.mark.parametrize(
    ["data", "pattern", "expected"],
    [
        # No broadcast
        (["1 2 3", "45 6", "789"], [r"\d+", r"\d+", r"\d"], [["1", "2", "3"], ["45", "6"], ["7", "8", "9"]]),
        # Broadcast pattern
        (["1 2 3", "45 6", "789"], [r"\d+"], [["1", "2", "3"], ["45", "6"], ["789"]]),
        # Broadcast data
        (["123"], [r"\d+", r"\d", r"\d+"], [["123"], ["1", "2", "3"], ["123"]]),
        # Broadcast null data
        ([None], [r"\d+", r"\d", r"\d+"], [None, None, None]),
        # Broadcast null pattern
        (["1 2 3", "45 6", "789"], [None], [None, None, None]),
        # Mixed in nulls
        (["1 2 3", None, "789"], [None, r"\d+", r"\d"], [None, None, ["7", "8", "9"]]),
    ],
)
def test_series_utf8_extract_all(data, pattern, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pattern, type=pa.string()))
    result = s.str.extract_all(patterns)
    assert result.to_pylist() == expected


@pytest.mark.parametrize("data", [["abc abc"]])
@pytest.mark.parametrize("pattern", [["(a)bc"]])
@pytest.mark.parametrize(
    "index, expected",
    [
        (0, [["abc", "abc"]]),
        (1, [["a", "a"]]),
        (2, [[]]),
    ],
)
def test_series_utf8_extract_all_index(data, pattern, index, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pattern, type=pa.string()))
    result = s.str.extract_all(patterns, index)
    assert result.to_pylist() == expected


def test_series_utf8_extract_all_bad_pattern() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pattern = Series.from_arrow(pa.array(["["]))
    with pytest.raises(ValueError):
        s.str.extract_all(pattern)


@pytest.mark.parametrize(
    ["data", "substrs", "expected"],
    [
        # No broadcast
        (["foo", "barbaz", "quux"], ["foo", "baz", "baz"], [0, 3, -1]),
        # Broadcast substrs
        (["foo", None, "quux"], ["foo"], [0, None, -1]),
        # Broadcast data
        (["foo"], ["foo", None, "baz"], [0, None, -1]),
        # Broadcast null data
        ([None], ["foo", "bar", "baz"], [None, None, None]),
        # Broadcast null substrs
        (["foo", "barbaz", "quux"], [None], [None, None, None]),
        # All Empty
        ([[], [], []]),
        # Mixed-in nulls
        (["foo", None, "barbaz", "quux"], ["oo", "bar", "baz", None], [1, None, 3, None]),
        # All null data.
        ([None] * 4, ["foo"] * 4, [None] * 4),
        # All null substrs
        (["foo"] * 4, [None] * 4, [None] * 4),
    ],
)
def test_series_utf8_find(data, substrs, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    substrs = Series.from_arrow(pa.array(substrs, type=pa.string()))
    result = s.str.find(substrs)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "substrs"],
    [
        # Empty substrs
        ([["foo"] * 4, []]),
    ],
)
def test_series_utf8_find_empty_arrs(data, substrs) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    substrs = Series.from_arrow(pa.array(substrs, type=pa.string()))
    with pytest.raises(ValueError):
        s.str.find(substrs)


def test_series_utf8_find_mismatch_len() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    substrs = Series.from_arrow(pa.array(["foo", "baz"], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.find(substrs)


def test_series_utf8_find_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    substrs = Series.from_arrow(pa.array(["foo", "baz", "quux"]))
    with pytest.raises(ValueError):
        s.str.find(substrs)


@pytest.mark.parametrize(
    ["data", "pattern", "replacement", "expected"],
    [
        # No broadcast
        (["foo", "barbaz", "quux"], ["o", "a", "u"], ["O", "A", "U"], ["fOO", "bArbAz", "qUUx"]),
        # Broadcast data
        (["foobar"], ["f", "o", "b"], ["F", "O", "B"], ["Foobar", "fOObar", "fooBar"]),
        # Broadcast pattern
        (["123", "12", "1"], ["1"], ["2", "3", "4"], ["223", "32", "4"]),
        # Broadcast replacement
        (["foo", "barbaz", "quux"], ["o", "a", "u"], [" "], ["f  ", "b rb z", "q  x"]),
        # Broadcast data and pattern
        (["foo"], ["o"], ["O", "A", "U"], ["fOO", "fAA", "fUU"]),
        # Broadcast data and replacement
        (["foo"], ["o", "f", " "], ["O"], ["fOO", "Ooo", "foo"]),
        # Broadcast pattern and replacement
        (["123", "12", "1"], ["1"], ["A"], ["A23", "A2", "A"]),
        # All empty
        ([], [], [], []),
    ],
)
@pytest.mark.parametrize("regex", [True, False])
def test_series_utf8_replace(data, pattern, replacement, expected, regex) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pattern, type=pa.string()))
    replacements = Series.from_arrow(pa.array(replacement, type=pa.string()))
    result = s.str.replace(patterns, replacements, regex=regex)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "pattern", "replacement", "expected"],
    [
        # Start of string.
        (["foo", "barbaz", "quux"], ["^f", "^b", "^q"], ["F", "B", "Q"], ["Foo", "Barbaz", "Quux"]),
        # Multi-character pattern.
        (["aabbaa", "abddaaaaaaa", "acaca"], [r"a+"], ["foo"], ["foobbfoo", "foobddfoo", "foocfoocfoo"]),
    ],
)
def test_series_utf8_replace_regex(data, pattern, replacement, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pattern, type=pa.string()))
    replacements = Series.from_arrow(pa.array(replacement, type=pa.string()))
    result = s.str.replace(patterns, replacements, regex=True)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "pattern", "replacement", "expected"],
    [
        # Broadcast null data
        ([None], ["o", "a", "u"], ["O", "A", "U"], [None, None, None]),
        # Broadcast null pattern
        (["foo", "barbaz", "quux"], [None], [" "], [None, None, None]),
        # Broadcast null replacement
        (["foo", "barbaz", "quux"], ["o", "a", "u"], [None], [None, None, None]),
        # Mixed-in nulls
        ([None, "barbaz", "quux", "1"], ["o", None, "u", "1"], ["O", "A", None, "2"], [None, None, None, "2"]),
    ],
)
@pytest.mark.parametrize("regex", [True, False])
def test_series_utf8_replace_nulls(data, pattern, replacement, expected, regex) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pattern, type=pa.string()))
    replacements = Series.from_arrow(pa.array(replacement, type=pa.string()))
    result = s.str.replace(patterns, replacements, regex=regex)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "pattern", "replacement"],
    [
        # Mismatched number of patterns and replacements
        (["foo", "barbaz", "quux"], ["o", "a"], ["O"]),
        (["foo", "barbaz", "quux"], [], ["O", "A"]),
        (["foo", "barbaz", "quux"], ["o", "a"], []),
        # bad input type
        ([1, 2, 3], ["o", "a"], ["O", "A"]),
    ],
)
@pytest.mark.parametrize("regex", [True, False])
def test_series_utf8_replace_bad_inputs(data, pattern, replacement, regex) -> None:
    s = Series.from_arrow(pa.array(data))
    with pytest.raises(ValueError):
        s.str.replace(pattern, replacement, regex=regex)

    pattern = Series.from_arrow(pa.array(pattern, type=pa.string()))
    replacement = Series.from_arrow(pa.array(replacement, type=pa.string()))
    with pytest.raises(ValueError):
        s.str.replace(pattern, replacement, regex=regex)


def test_series_utf8_replace_bad_regex_pattern() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pattern = Series.from_arrow(pa.array(["["]))
    replacement = Series.from_arrow(pa.array([" "]))
    with pytest.raises(ValueError):
        s.str.replace(pattern, replacement, regex=True)


@pytest.mark.parametrize(
    ["data", "length", "pad", "expected"],
    [
        # No Broadcast
        (["foo", "barbaz", "quux"], [5, 6, 7], [" ", ".", "-"], ["foo  ", "barbaz", "quux---"]),
        (["foo", "barbaz", "quux"], [5, 6, 7], ["-", ".", " "], ["foo--", "barbaz", "quux   "]),
        # Broadcast length
        (["foo", "barbaz", "quux"], [5], ["-", ".", " "], ["foo--", "barba", "quux "]),
        # Broadcast data
        (["foo"], [5, 6, 7], ["-", ".", " "], ["foo--", "foo...", "foo    "]),
        # Broadcast pad
        (["foo", "barbaz", "quux"], [5, 6, 7], ["-"], ["foo--", "barbaz", "quux---"]),
        # Broadcast null data
        ([None], [5, 6, 7], ["-", ".", "-"], [None, None, None]),
        # Broadcast null length
        (["foo", "barbaz", "quux"], [None], ["-", ".", " "], [None, None, None]),
        # Broadcast null pad
        (["foo", "barbaz", "quux"], [5, 6, 7], [None], [None, None, None]),
        # All Empty
        ([[], [], [], []]),
        # Mixed-in nulls
        (["foo", None, "barbaz", "quux"], [5, 6, 7, 8], ["-", ".", " ", None], ["foo--", None, "barbaz ", None]),
        # All null data.
        ([None] * 4, [5] * 4, ["-", ".", " ", "_"], [None] * 4),
        # All null length
        (["foo"] * 4, [None] * 4, ["-", ".", " ", "_"], [None] * 4),
        # All null pad
        (["foo"] * 4, [5] * 4, [None] * 4, [None] * 4),
        # With emojis
        (["😃😌😝", "abc"], [5, 6], ["😅"], ["😃😌😝😅😅", "abc😅😅😅"]),
    ],
)
def test_series_utf8_rpad(data, length, pad, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    lengths = Series.from_arrow(pa.array(length, type=pa.uint32()))
    pads = Series.from_arrow(pa.array(pad, type=pa.string()))
    result = s.str.rpad(lengths, pads)
    assert result.to_pylist() == expected


def test_series_utf8_rpad_mismatch_len() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    lengths = Series.from_arrow(pa.array([5, 6], type=pa.uint32()))
    pads = Series.from_arrow(pa.array(["-", "."], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.rpad(lengths, pads)


def test_series_utf8_rpad_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    lengths = Series.from_arrow(pa.array([5, 6, 7], type=pa.uint32()))
    pads = Series.from_arrow(pa.array(["-", ".", " "], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.rpad(lengths, pads)


def test_series_utf8_rpad_bad_lengths_dtype() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    lengths = Series.from_arrow(pa.array(["5", "6", "7"]))
    pads = Series.from_arrow(pa.array(["-", ".", " "], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.rpad(lengths, pads)


def test_series_utf8_rpad_bad_pads_dtype() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    lengths = Series.from_arrow(pa.array([5, 6, 7], type=pa.uint32()))
    pads = Series.from_arrow(pa.array([1, 2, 3]))
    with pytest.raises(ValueError):
        s.str.rpad(lengths, pads)


def test_series_utf8_rpad_multichar_pad() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    lengths = Series.from_arrow(pa.array([5, 6, 7], type=pa.uint32()))
    pads = Series.from_arrow(pa.array(["--", "..", "  "], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.rpad(lengths, pads)


@pytest.mark.parametrize(
    ["data", "length", "pad"],
    [
        # empty length
        (["foo", "barbaz", "quux"], [], ["-", ".", " "]),
        # empty pad
        (["foo", "barbaz", "quux"], [5, 6, 7], []),
    ],
)
def test_series_utf8_rpad_empty_arrs(data, length, pad) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    lengths = Series.from_arrow(pa.array(length, type=pa.uint32()))
    pads = Series.from_arrow(pa.array(pad, type=pa.string()))
    with pytest.raises(ValueError):
        s.str.rpad(lengths, pads)


@pytest.mark.parametrize(
    ["data", "length", "pad", "expected"],
    [
        # No Broadcast
        (["foo", "barbaz", "quux"], [5, 6, 7], [" ", ".", "-"], ["  foo", "barbaz", "---quux"]),
        (["foo", "barbaz", "quux"], [5, 4, 7], ["-", ".", " "], ["--foo", "barb", "   quux"]),
        # Broadcast length
        (["foo", "barbaz", "quux"], [5], ["-", ".", " "], ["--foo", "barba", " quux"]),
        # Broadcast data
        (["foo"], [5, 6, 7], ["-", ".", " "], ["--foo", "...foo", "    foo"]),
        # Broadcast pad
        (["foo", "barbaz", "quux"], [5, 4, 7], ["-"], ["--foo", "barb", "---quux"]),
        # Broadcast null data
        ([None], [5, 6, 7], ["-", ".", "-"], [None, None, None]),
        # Broadcast null length
        (["foo", "barbaz", "quux"], [None], ["-", ".", " "], [None, None, None]),
        # Broadcast null pad
        (["foo", "barbaz", "quux"], [5, 6, 7], [None], [None, None, None]),
        # All Empty
        ([[], [], [], []]),
        # Mixed-in nulls
        (["foo", None, "barbaz", "quux"], [5, 6, 7, 8], ["-", ".", " ", None], ["--foo", None, " barbaz", None]),
        # All null data.
        ([None] * 4, [5] * 4, ["-", ".", " ", "_"], [None] * 4),
        # All null length
        (["foo"] * 4, [None] * 4, ["-", ".", " ", "_"], [None] * 4),
        # All null pad
        (["foo"] * 4, [5] * 4, [None] * 4, [None] * 4),
        # With emojis
        (["😃😌😝", "abc"], [4, 6], ["😅"], ["😅😃😌😝", "😅😅😅abc"]),
    ],
)
def test_series_utf8_lpad(data, length, pad, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    lengths = Series.from_arrow(pa.array(length, type=pa.uint32()))
    pads = Series.from_arrow(pa.array(pad, type=pa.string()))
    result = s.str.lpad(lengths, pads)
    assert result.to_pylist() == expected


def test_series_utf8_lpad_mismatch_len() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    lengths = Series.from_arrow(pa.array([5, 6], type=pa.uint32()))
    pads = Series.from_arrow(pa.array(["-", "."], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.lpad(lengths, pads)


def test_series_utf8_lpad_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    lengths = Series.from_arrow(pa.array([5, 6, 7], type=pa.uint32()))
    pads = Series.from_arrow(pa.array(["-", ".", " "], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.lpad(lengths, pads)


def test_series_utf8_lpad_bad_lengths_dtype() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    lengths = Series.from_arrow(pa.array(["5", "6", "7"]))
    pads = Series.from_arrow(pa.array(["-", ".", " "], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.lpad(lengths, pads)


def test_series_utf8_lpad_bad_pads_dtype() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    lengths = Series.from_arrow(pa.array([5, 6, 7], type=pa.uint32()))
    pads = Series.from_arrow(pa.array([1, 2, 3]))
    with pytest.raises(ValueError):
        s.str.lpad(lengths, pads)


def test_series_utf8_lpad_multichar_pad() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    lengths = Series.from_arrow(pa.array([5, 6, 7], type=pa.uint32()))
    pads = Series.from_arrow(pa.array(["--", "..", "  "], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.lpad(lengths, pads)


@pytest.mark.parametrize(
    ["data", "length", "pad"],
    [
        # empty length
        (["foo", "barbaz", "quux"], [], ["-", ".", " "]),
        # empty pad
        (["foo", "barbaz", "quux"], [5, 6, 7], []),
    ],
)
def test_series_utf8_lpad_empty_arrs(data, length, pad) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    lengths = Series.from_arrow(pa.array(length, type=pa.uint32()))
    pads = Series.from_arrow(pa.array(pad, type=pa.string()))
    with pytest.raises(ValueError):
        s.str.lpad(lengths, pads)


@pytest.mark.parametrize(
    ["data", "n", "expected"],
    [
        # No Broadcast
        (["foo", "barbaz", "quux"], [0, 1, 2], ["", "barbaz", "quuxquux"]),
        # Broadcast n
        (["foo", "barbaz", "quux"], [1], ["foo", "barbaz", "quux"]),
        # Broadcast data
        (["foo"], [0, 1, 2], ["", "foo", "foofoo"]),
        # Broadcast null data
        ([None], [0, 1, 2], [None, None, None]),
        # Broadcast null n
        (["foo", "barbaz", "quux"], [None], [None, None, None]),
        # All Empty
        ([[], [], []]),
        # Mixed-in nulls
        (["foo", None, "barbaz", "quux"], [0, 1, 1, None], ["", None, "barbaz", None]),
        # All null data.
        ([None] * 4, [1] * 4, [None] * 4),
        # All null n
        (["foo"] * 4, [None] * 4, [None] * 4),
        # with emojis
        (["😃😌😝", "abc😃😄😅"], [3, 2], ["😃😌😝😃😌😝😃😌😝", "abc😃😄😅abc😃😄😅"]),
    ],
)
def test_series_utf8_repeat(data, n, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    ntimes = Series.from_arrow(pa.array(n, type=pa.uint32()))
    result = s.str.repeat(ntimes)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    ["data", "n"],
    [
        # empty n
        ([["foo"] * 4, []]),
    ],
)
def test_series_utf8_repeat_empty_arrs(data, n) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    n = Series.from_arrow(pa.array(n, type=pa.uint32()))
    with pytest.raises(ValueError):
        s.str.repeat(n)


def test_series_utf8_repeat_mismatch_len() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    n = Series.from_arrow(pa.array([1, 2], type=pa.uint32()))
    with pytest.raises(ValueError):
        s.str.repeat(n)


def test_series_utf8_repeat_bad_ntype() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    n = Series.from_arrow(pa.array(["1", "2", "3"]))
    with pytest.raises(ValueError):
        s.str.repeat(n)


def test_series_utf8_repeat_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    n = Series.from_arrow(pa.array([1, 2, 3]))
    with pytest.raises(ValueError):
        s.str.repeat(n)


def test_series_utf8_like_bad_pattern() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pattern = Series.from_arrow(pa.array(["["]))
    with pytest.raises(ValueError):
        s.str.like(pattern)


def test_series_utf8_like_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    pattern = Series.from_arrow(pa.array(["foo", "baz", "quux"]))
    with pytest.raises(ValueError):
        s.str.like(pattern)


def test_series_utf8_like_mismatch_len() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pattern = Series.from_arrow(pa.array(["foo", "baz"], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.like(pattern)


@pytest.mark.parametrize(
    ["data", "pat", "expected"],
    [
        # No Broadcast
        (["foo", "barbaz", "quux"], ["foo", "%baz", "Quu_"], [True, True, False]),
        # Broadcast pattern
        (["foo", "barbaz", "quux"], ["f_"], [False, False, False]),
        # Broadcast data
        (["foo"], ["foo", "%baz", "quu_"], [True, False, False]),
        # Broadcast null data
        ([None], ["foo", "%baz", "quu_"], [None, None, None]),
        # Broadcast null pattern
        (["foo", "barbaz", "quux"], [None], [None, None, None]),
        # Mixed-in nulls
        (["foo", None, "barbaz", "quux"], ["foo", "quu_", "%baz", None], [True, None, True, None]),
        # All null data.
        ([None] * 4, ["foo", "%baz", "quu_", None], [None] * 4),
        # All null pattern
        (["foo"] * 4, [None], [None] * 4),
    ],
)
def test_series_utf8_like(data, pat, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pat, type=pa.string()))
    result = s.str.like(patterns)
    assert result.to_pylist() == expected


def test_series_utf8_like_empty_arrs() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pat = Series.from_arrow(pa.array([], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.like(pat)


def test_series_utf8_ilike_bad_pattern() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pattern = Series.from_arrow(pa.array(["["]))
    with pytest.raises(ValueError):
        s.str.ilike(pattern)


def test_series_utf8_ilike_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    pattern = Series.from_arrow(pa.array(["foo", "baz", "quux"]))
    with pytest.raises(ValueError):
        s.str.ilike(pattern)


def test_series_utf8_ilike_mismatch_len() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pattern = Series.from_arrow(pa.array(["foo", "baz"], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.ilike(pattern)


@pytest.mark.parametrize(
    ["data", "pat", "expected"],
    [
        # No Broadcast
        (["foo", "barbaz", "quux"], ["foo", "%baz", "Quu_"], [True, True, True]),
        # Broadcast pattern
        (["foo", "barbaz", "quux"], ["f_"], [False, False, False]),
        # Broadcast data
        (["foo"], ["foo", "%baz", "quu_"], [True, False, False]),
        # Broadcast null data
        ([None], ["foo", "%baz", "quu_"], [None, None, None]),
        # Broadcast null pattern
        (["foo", "barbaz", "quux"], [None], [None, None, None]),
        # Mixed-in nulls
        (["foo", None, "barbaz", "quux"], ["foo", "quu_", "%baz", None], [True, None, True, None]),
        # All null data.
        ([None] * 4, ["foo", "%baz", "quu_", None], [None] * 4),
        # All null pattern
        (["foo"] * 4, [None], [None] * 4),
    ],
)
def test_series_utf8_ilike(data, pat, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    patterns = Series.from_arrow(pa.array(pat, type=pa.string()))
    result = s.str.ilike(patterns)
    assert result.to_pylist() == expected


def test_series_utf8_ilike_empty_arrs() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    pat = Series.from_arrow(pa.array([], type=pa.string()))
    with pytest.raises(ValueError):
        s.str.ilike(pat)


@pytest.mark.parametrize(
    ["data", "start", "length", "expected"],
    [
        pytest.param(["foo", "barbaz", "quux"], [0, 1, 2], [1, 1, 1], ["f", "a", "u"], id="No broadcast"),
        pytest.param(["foo", "barbaz", "quux"], [0], [1, 1, 1], ["f", "b", "q"], id="Broadcast start"),
        pytest.param(["foo", "barbaz", "quux"], [1, 1, 1], [2], ["oo", "ar", "uu"], id="Broadcast length"),
        pytest.param(["foo"], [0, 0, 0], [1, 2, 3], ["f", "fo", "foo"], id="Broadcast data"),
        pytest.param([None], [0, 1, 2], [0, 1, 2], [None, None, None], id="Broadcast null data"),
        pytest.param(["foo", "barbaz", "quux"], [None], [1, 1, 1], [None, None, None], id="Broadcast null start"),
        pytest.param(
            ["foo", "barbaz", "quux"], [0, 0, 0], [None], ["foo", "barbaz", "quux"], id="Broadcast null length"
        ),
        pytest.param([], [], [], [], id="All empty"),
        pytest.param([None] * 4, [1] * 4, [1] * 4, [None] * 4, id="All null data"),
        pytest.param(["foo"] * 4, [None] * 4, [1] * 4, [None] * 4, id="All null start"),
        pytest.param(["foo"] * 4, [0] * 4, [None] * 4, ["foo"] * 4, id="All null length"),
        pytest.param(["foo"] * 4, [None] * 4, [None] * 4, [None] * 4, id="All null length and length"),
        pytest.param(["😃😌😝", "abc😃😄😅"], [1, 3], [None], ["😌😝", "😃😄😅"], id="With emojis"),
        pytest.param(["foo"], [0], [0], [None], id="Zero length"),
        pytest.param(["foo"], [5], [10], [None], id="Start over the string length"),
        pytest.param(["foo", "bar"], [0, 1], [None, None], ["foo", "ar"], id="None series length"),
    ],
)
def test_series_utf8_substr(data, start, length, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    start = Series.from_arrow(pa.array(start, type=pa.uint32()))
    length = Series.from_arrow(pa.array(length, type=pa.uint32()))
    result = s.str.substr(start, length)
    assert result.to_pylist() == expected


def test_series_utf8_substr_length_is_none() -> None:
    s = Series.from_arrow(pa.array(["foo", "bar", "baz"], type=pa.string()))
    start = Series.from_arrow(pa.array([0, 1, 2], type=pa.uint32()))
    result = s.str.substr(start, None)
    assert result.to_pylist() == ["foo", "ar", "z"]


@pytest.mark.parametrize(
    ("data", "start", "length"),
    [
        pytest.param(["foo"] * 4, [], [1] * 4, id="Empty start"),
        pytest.param(["foo"] * 4, [] * 4, [], id="Empty length"),
    ],
)
def test_series_utf8_substr_empty_arrs(data, start, length) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    start = Series.from_arrow(pa.array(start, type=pa.uint32()))
    length = Series.from_arrow(pa.array(length, type=pa.uint32()))

    with pytest.raises(ValueError):
        s.str.substr(start, length)


def test_series_utf8_substr_mismatch_start() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    start = Series.from_arrow(pa.array([1, 2], type=pa.uint32()))
    with pytest.raises(ValueError):
        s.str.substr(start)


def test_series_utf8_substr_mismatch_length() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    start = Series.from_arrow(pa.array([1, 2, 3], type=pa.uint32()))
    length = Series.from_arrow(pa.array([1, 2], type=pa.uint32()))
    with pytest.raises(ValueError):
        s.str.substr(start, length)


def test_series_utf8_substr_bad_start_type() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    start = Series.from_arrow(pa.array(["1", "2", "3"]))
    with pytest.raises(ValueError):
        s.str.substr(start)


def test_series_utf8_substr_bad_length_type() -> None:
    s = Series.from_arrow(pa.array(["foo", "barbaz", "quux"]))
    length = Series.from_arrow(pa.array(["1", "2", "3"]))
    with pytest.raises(ValueError):
        s.str.substr(0, length)


def test_series_utf8_substr_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    start = Series.from_arrow(pa.array([1, 2, 3]))
    with pytest.raises(ValueError):
        s.str.substr(start)


@pytest.mark.parametrize(
    ["data", "format", "expected"],
    [
        (
            ["2021-01-01", "2021-01-02", "2021-01-03"],
            "%Y-%m-%d",
            [datetime.date(2021, 1, 1), datetime.date(2021, 1, 2), datetime.date(2021, 1, 3)],
        ),
        (
            # null data
            [None],
            "%Y-%m-%d",
            [None],
        ),
        (
            # mixed-in nulls
            ["2021-01-01", None, "2021-01-03"],
            "%Y-%m-%d",
            [datetime.date(2021, 1, 1), None, datetime.date(2021, 1, 3)],
        ),
        (
            # all empty
            [],
            "%Y-%m-%d",
            [],
        ),
    ],
)
def test_series_utf8_to_date(data, format, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    result = s.str.to_date(format)
    assert result.to_pylist() == expected


def test_series_utf8_to_date_bad_format() -> None:
    s = Series.from_arrow(pa.array(["2021-01-01", "2021-01-02"]))
    format = "%Y-%m-%"
    with pytest.raises(ValueError):
        s.str.to_date(format)


def test_series_utf8_to_date_bad_format_type() -> None:
    s = Series.from_arrow(pa.array(["2021-01-01", "2021-01-02"]))
    format = 1
    with pytest.raises(ValueError):
        s.str.to_date(format)


def test_series_utf8_to_date_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    format = "%Y-%m-%d"
    with pytest.raises(ValueError):
        s.str.to_date(format)


def test_series_utf8_bad_date() -> None:
    s = Series.from_arrow(pa.array(["2021-100-20"]))
    format = "%Y-%m-%d"
    with pytest.raises(ValueError):
        s.str.to_date(format)


@pytest.mark.parametrize(
    ["data", "format", "timezone", "expected"],
    [
        pytest.param(
            ["2021-01-01 00:00:00", "2021-01-02 01:07:35", "2021-01-03 12:30:00"],
            "%Y-%m-%d %H:%M:%S",
            None,
            [
                datetime.datetime(2021, 1, 1, 0, 0),
                datetime.datetime(2021, 1, 2, 1, 7, 35),
                datetime.datetime(2021, 1, 3, 12, 30),
            ],
            id="No timezone",
        ),
        pytest.param(
            ["2021-01-01 00:00:00 +0530", "2021-01-02 01:07:35 +0530", "2021-01-03 12:30:00 +0530"],
            "%Y-%m-%d %H:%M:%S %z",
            "Asia/Kolkata",
            [
                datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone(datetime.timedelta(hours=5, minutes=30))),
                datetime.datetime(
                    2021, 1, 2, 1, 7, 35, tzinfo=datetime.timezone(datetime.timedelta(hours=5, minutes=30))
                ),
                datetime.datetime(
                    2021, 1, 3, 12, 30, tzinfo=datetime.timezone(datetime.timedelta(hours=5, minutes=30))
                ),
            ],
            id="IST timezone",
        ),
        pytest.param(
            ["2021-01-01 00:00:00 +0000", "2021-01-02 01:07:35 +0100", "2021-01-03 12:30:00 +0200"],
            "%Y-%m-%d %H:%M:%S %z",
            "UTC",
            [
                datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc),
                datetime.datetime(2021, 1, 2, 0, 7, 35, tzinfo=datetime.timezone.utc),
                datetime.datetime(2021, 1, 3, 10, 30, tzinfo=datetime.timezone.utc),
            ],
            id="Different timezones",
        ),
        pytest.param(
            ["2021-01-01 00:00:00 +0000", "2021-01-02 01:07:35 +0100", "2021-01-03 12:30:00 +0200"],
            "%Y-%m-%d %H:%M:%S %z",
            None,
            [
                datetime.datetime(2021, 1, 1, 0, 0, tzinfo=datetime.timezone.utc),
                datetime.datetime(2021, 1, 2, 0, 7, 35, tzinfo=datetime.timezone.utc),
                datetime.datetime(2021, 1, 3, 10, 30, tzinfo=datetime.timezone.utc),
            ],
            id="Unspecified timezone returns datetime in utc",
        ),
    ],
)
def test_series_utf8_to_datetime_different_timezones(data, format, timezone, expected) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    result = s.str.to_datetime(format, timezone)
    assert result.to_pylist() == expected


@pytest.mark.parametrize(
    "data,format,expected,timeunit",
    [
        pytest.param(
            ["2021-01-01 00:00:00", "2021-01-02 01:07:35", "2021-01-03 12:30:00"],
            "%Y-%m-%d %H:%M:%S",
            [
                datetime.datetime(2021, 1, 1, 0, 0),
                datetime.datetime(2021, 1, 2, 1, 7, 35),
                datetime.datetime(2021, 1, 3, 12, 30),
            ],
            "us",
            id="Default to Microseconds",
        ),
        pytest.param(
            ["2021-01-01 00:00:45.123456", "2021-01-02 01:07:35.123456", "2021-01-03 12:30:00.123456"],
            "%Y-%m-%d %H:%M:%S%.f",
            [
                datetime.datetime(2021, 1, 1, 0, 0, 45, 123456),
                datetime.datetime(2021, 1, 2, 1, 7, 35, 123456),
                datetime.datetime(2021, 1, 3, 12, 30, 0, 123456),
            ],
            "us",
            id="Microseconds",
        ),
        pytest.param(
            ["2021-01-01 00:00:45.1234", "2021-01-02 01:07:35.12300", "2021-01-03 12:30:00.123"],
            "%Y-%m-%d %H:%M:%S%.3f",
            [
                datetime.datetime(2021, 1, 1, 0, 0, 45, 123000),
                datetime.datetime(2021, 1, 2, 1, 7, 35, 123000),
                datetime.datetime(2021, 1, 3, 12, 30, 0, 123000),
            ],
            "ms",
            id="Milliseconds",
        ),
    ],
)
def test_series_utf8_to_datetime_different_timeunit(data, format, expected, timeunit) -> None:
    s = Series.from_arrow(pa.array(data, type=pa.string()))
    result = s.str.to_datetime(format)
    assert result.to_pylist() == expected
    assert result.datatype() == DataType.timestamp(timeunit)


def test_series_utf8_to_datetime_bad_format() -> None:
    s = Series.from_arrow(pa.array(["2021-01-01", "2021-01-02"]))
    format = "%Y-%m-%"
    with pytest.raises(ValueError):
        s.str.to_datetime(format, "UTC")


def test_series_utf8_to_datetime_bad_format_type() -> None:
    s = Series.from_arrow(pa.array(["2021-01-01", "2021-01-02"]))
    format = 1
    with pytest.raises(ValueError):
        s.str.to_datetime(format, "UTC")


def test_series_to_datetime_bad_timezone_dtype() -> None:
    s = Series.from_arrow(pa.array(["2021-01-01", "2021-01-02"]))
    with pytest.raises(ValueError):
        s.str.to_datetime("%Y-%m-%d %H:%M:%S", 1)


def test_series_to_datetime_bad_timezone() -> None:
    s = Series.from_arrow(pa.array(["2021-01-01", "2021-01-02"]))
    with pytest.raises(ValueError):
        s.str.to_datetime("%Y-%m-%d %H:%M:%S", "foo")


def test_series_utf8_to_datetime_bad_dtype() -> None:
    s = Series.from_arrow(pa.array([1, 2, 3]))
    with pytest.raises(ValueError):
        s.str.to_datetime("%Y-%m-%d %H:%M:%S", "UTC")


def test_series_utf8_to_bad_datetime() -> None:
    s = Series.from_arrow(pa.array(["2021-100-20"]))
    with pytest.raises(ValueError):
        s.str.to_datetime("%Y-%m-%d %H:%M:%S", "UTC")


# source: RedPajama
def manual_normalize(text, remove_punct, lowercase, nfd_unicode, white_space):
    if text is None:
        return None

    if remove_punct:
        text = text.translate(str.maketrans("", "", string.punctuation))

    if lowercase:
        text = text.lower()

    if white_space:
        text = text.strip()
        text = re.sub(r"\s+", " ", text)

    if nfd_unicode:
        text = unicodedata.normalize("NFD", text)

    return text


NORMALIZE_TEST_DATA = [
    "regular text no changes",
    "Regular texT WITH uPpErCaSe",
    "ALL UPPERCASE TEXT",
    "text, with... punctuation!!!",
    "!&# #%*!()%*@# &*%#& @*( #*(@%()))",
    "!@#$%^&*()+_-=~`[]\\/.,';?><\":|}{",
    "UPPERCASE, AND, PUNCTUATION!?",
    "füñķÿ úňìčõðė",
    "füñķÿ, úňìčõðė!",
    "FüÑķŸ úňÌčõÐė",
    "FüÑķŸ,   úňÌčõÐė!",
    "way    too    much      space",
    "     space  all     over   the place    ",
    "other\ntypes\tof\r\nspace characters",
    "too\n\n\t\r\n  \n\t\tmuch\n\t\tspace\t  \n\n  \t\r\n \t",
    None,
    "TOO\n\n\t\r\n  \n\t\tMUCH\n\t\tsPACe\t  \n\n  \t\r\n \t",
    "too,\n\n?\t\r\n  \n\t\tmuc%h!!%\n\t\t\\SPACE??!\t  \n\n  \t\r\n \t",
    "FüÑķŸ,   úňÌčõÐė!   AND EVERYTHING else TOO    \t\na\t\t\nbCDe 😃😌😝",
    "",
    "specialcase",
    "SPECIALCASE",
    "😃",
    None,
]


@pytest.mark.parametrize("remove_punct", [False, True])
@pytest.mark.parametrize("lowercase", [False, True])
@pytest.mark.parametrize("nfd_unicode", [False, True])
@pytest.mark.parametrize("white_space", [False, True])
def test_series_utf8_normalize(remove_punct, lowercase, nfd_unicode, white_space) -> None:
    s = Series.from_pylist(NORMALIZE_TEST_DATA)
    a = s.str.normalize(
        remove_punct=remove_punct,
        lowercase=lowercase,
        nfd_unicode=nfd_unicode,
        white_space=white_space,
    ).to_pylist()
    b = [manual_normalize(t, remove_punct, lowercase, nfd_unicode, white_space) for t in NORMALIZE_TEST_DATA]
    assert a == b


def test_series_utf8_count_matches():
    s = Series.from_pylist(
        [
            "the quick brown fox jumped over the lazy dog",
            "the quick brown foe jumped o'er the lazy dot",
            "the fox fox fox jumped over over dog lazy dog",
            "the quick brown foxes hovered above the lazy dogs",
            "the quick brown-fox jumped over the 'lazy dog'",
            "thequickbrownfoxjumpedoverthelazydog",
            "THE QUICK BROWN FOX JUMPED over THE Lazy DOG",
            "    fox     dog        over    ",
        ]
    )
    p = Series.from_pylist(
        [
            "fox",
            "over",
            "lazy dog",
            "dog",
        ]
    )

    res = s.str.count_matches(p, whole_words=False, case_sensitive=False).to_pylist()
    assert res == [3, 0, 7, 3, 3, 3, 3, 3]
    res = s.str.count_matches(p, whole_words=True, case_sensitive=False).to_pylist()
    assert res == [3, 0, 7, 0, 3, 0, 3, 3]
    res = s.str.count_matches(p, whole_words=False, case_sensitive=True).to_pylist()
    assert res == [3, 0, 7, 3, 3, 3, 1, 3]
    res = s.str.count_matches(p, whole_words=True, case_sensitive=True).to_pylist()
    assert res == [3, 0, 7, 0, 3, 0, 1, 3]


@pytest.mark.parametrize("whole_words", [False, True])
@pytest.mark.parametrize("case_sensitive", [False, True])
def test_series_utf8_count_matches_overlap(whole_words, case_sensitive):
    s = Series.from_pylist(["hello world"])
    p = Series.from_pylist(["hello world", "hello", "world"])
    res = s.str.count_matches(p, whole_words=whole_words, case_sensitive=case_sensitive).to_pylist()
    assert res == [1]



from __future__ import annotations


def test_with_column(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    expanded_df = df.with_column("bar", df["sepal_width"] + df["petal_length"])
    data = expanded_df.to_pydict()
    assert expanded_df.column_names == df.column_names + ["bar"]
    assert data["bar"] == [sw + pl for sw, pl in zip(data["sepal_width"], data["petal_length"])]


def test_with_column_same_name(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    expanded_df = df.with_column("sepal_width", df["sepal_width"] + df["petal_length"])
    data = expanded_df.to_pydict()
    assert expanded_df.column_names == df.column_names
    expected = [valid_data[i]["sepal_width"] + valid_data[i]["petal_length"] for i in range(len(valid_data))]
    assert data["sepal_width"] == expected


def test_stacked_with_columns(make_df, valid_data: list[dict[str, float]]):
    df = make_df(valid_data)
    df = df.select(df["sepal_length"])
    df = df.with_column("sepal_length_2", df["sepal_length"])
    df = df.with_column("sepal_length_3", df["sepal_length_2"])
    pdf = df.to_pandas()
    assert len(pdf.columns) == 3
    assert pdf.columns.to_list() == ["sepal_length", "sepal_length_2", "sepal_length_3"]



"""Tests for HTML representation of Series and array values."""

from __future__ import annotations

import pytest

import daft


def test_utf8_html_value_truncation():
    """Test that strings longer than 1MB are truncated in HTML representation."""
    # Create a string longer than 1MB
    long_string = "a" * (1024 * 1024 + 100)  # 1MB + 100 bytes
    df = daft.from_pydict({"test_col": [long_string]})

    # Get the HTML representation
    html_repr = df._repr_html_()

    # The actual content should be truncated - check that the long string doesn't appear in full
    # The HTML table structure adds overhead, so we check for truncation indicator instead
    assert "..." in html_repr

    # Verify the original long string isn't fully present in the HTML
    assert long_string not in html_repr


def test_utf8_html_value_short_string():
    """Test that short strings are not truncated."""
    short_string = "Hello, world!"
    df = daft.from_pydict({"test_col": [short_string]})

    html_repr = df._repr_html_()

    # Should contain the original string
    assert "Hello, world!" in html_repr
    # Should not contain truncation indicator (except for possible table formatting)
    assert html_repr.count("...") <= 1  # Allow for table formatting


def test_utf8_html_value_html_escaping():
    """Test that HTML special characters are properly escaped."""
    html_string = "<script>alert('xss')</script>"
    df = daft.from_pydict({"test_col": [html_string]})

    html_repr = df._repr_html_()

    # Should be HTML escaped
    assert "&lt;script&gt;" in html_repr
    assert "&lt;/script&gt;" in html_repr
    # Should not contain raw HTML
    assert "<script>" not in html_repr


def test_utf8_html_value_newline_replacement():
    """Test that newlines are replaced with <br /> tags."""
    multiline_string = "line1\nline2\nline3"
    df = daft.from_pydict({"test_col": [multiline_string]})

    html_repr = df._repr_html_()

    # Should replace newlines with <br />
    assert "<br />" in html_repr


def test_unicode_boundary_handling():
    """Test that truncation respects UTF-8 character boundaries."""
    # Create a string with unicode characters near the 1MB boundary
    unicode_char = "🦀"  # 4-byte UTF-8 character
    base_string = "a" * (1024 * 1024 - 2)  # Just under 1MB
    long_string = base_string + unicode_char

    df = daft.from_pydict({"test_col": [long_string]})
    html_repr = df._repr_html_()

    # Should not crash (which would happen if we split in the middle of a UTF-8 character)
    assert isinstance(html_repr, str)
    assert len(html_repr) > 0


def test_python_object_html_representation():
    """Test that Python objects with large string representations are truncated."""

    class LargeObject:
        def __str__(self):
            return "large_data:" + "x" * (1024 * 1024 + 100)  # 1MB + 100 bytes

    large_obj = LargeObject()
    # Create a dataframe with the Python object
    df = daft.from_pydict({"test_col": [large_obj]})

    html_repr = df._repr_html_()

    # Should not crash and should be truncated
    assert isinstance(html_repr, str)
    # The full object string shouldn't be present
    full_obj_str = str(large_obj)
    assert full_obj_str not in html_repr


def test_list_html_representation():
    """Test that lists with large string representations are truncated."""
    # Create a list with many elements that would produce a large string
    large_list = ["item_" + str(i) for i in range(10000)]  # Reduced size for faster testing
    df = daft.from_pydict({"test_col": [large_list]})

    html_repr = df._repr_html_()

    # Should not crash and should be reasonably sized
    assert isinstance(html_repr, str)
    # Should be much smaller than the full representation would be
    assert len(html_repr) < 1024 * 1024 * 2  # Allow some overhead for HTML structure


@pytest.mark.parametrize("data_type", ["string", "binary"])
def test_various_data_types_truncation(data_type):
    """Test that various data types handle large representations correctly."""
    if data_type == "string":
        large_data = "x" * (1024 * 1024 + 100)
        df = daft.from_pydict({"test_col": [large_data]})
    elif data_type == "binary":
        large_data = b"x" * (1024 * 1024 + 100)
        df = daft.from_pydict({"test_col": [large_data]})

    html_repr = df._repr_html_()

    # Should not crash and should be truncated
    assert isinstance(html_repr, str)
    assert len(html_repr) < 1024 * 1024 * 2  # Allow some overhead



from __future__ import annotations

import pytest

import daft


def test_limit():
    df = daft.from_pydict({"id": [i for i in range(1024)], "name": [f"user_{i}" for i in range(1024)]})

    df = df.select("name").limit(1)
    pdf = df.to_pandas()

    assert len(pdf) == 1, f"Expected 1 rows, got {len(pdf)}"
    assert pdf["name"].iloc[0] == "user_0", f"Expected 'user_0', got {pdf['name'].iloc[0]}"


def test_negative_limit():
    df = daft.from_pydict({"id": [i for i in range(1024)], "name": [f"user_{i}" for i in range(1024)]})

    with pytest.raises(ValueError) as excinfo:
        df.select("name").limit(-1)

    assert "LIMIT <n> must be greater than or equal to 0, instead got: -1" in str(excinfo.value)



from __future__ import annotations

import pytest

PATH = (
    "s3://eventual-dev-benchmarking-fixtures/parquet-benchmarking/tpch/200MB-2RG/daft_200MB_lineitem_chunk.RG-2.parquet"
)


@pytest.mark.benchmark(group="num_files_single_column")
@pytest.mark.parametrize(
    "num_files",
    [1, 2, 4, 8],
)
def test_read_parquet_num_files_single_column(num_files, bulk_read_fn, benchmark):
    data = benchmark(bulk_read_fn, [PATH] * num_files, columns=["L_ORDERKEY"])
    assert len(data) == num_files
    # Make sure the data is correct
    for i in range(num_files):
        assert data[i].column_names == ["L_ORDERKEY"]
        assert len(data[i]) == 5515199


@pytest.mark.benchmark(group="num_rowgroups_all_columns")
@pytest.mark.parametrize(
    "num_files",
    [1, 2, 4],
)
def test_read_parquet_num_files_all_columns(num_files, bulk_read_fn, benchmark):
    data = benchmark(bulk_read_fn, [PATH] * num_files)
    assert len(data) == num_files

    # Make sure the data is correct
    for i in range(num_files):
        assert len(data[i].column_names) == 16
        assert len(data[i]) == 5515199



from __future__ import annotations

import daft
from daft import col


def test_url_parse_basic():
    urls = [
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.ar/train-00004-of-00007.parquet",
        "https://user:pass@example.com:8080/path?query=value#fragment",
        "http://localhost/api",
        "ftp://files.example.com/file.txt",
    ]

    df = daft.from_pydict({"urls": urls})

    result = df.select(col("urls").url_parse()).collect()

    assert len(result) == 4
    schema = result.schema()
    assert len(schema) == 1
    field = schema["urls"]
    assert field.name == "urls"
    assert field.dtype.is_struct()

    result_data = result.to_pydict()
    assert len(result_data["urls"]) == 4
    assert all(isinstance(url_struct, dict) for url_struct in result_data["urls"] if url_struct is not None)


def test_url_parse_component_extraction():
    urls = [
        "https://user:pass@example.com:8080/path?query=value#fragment",
        "http://localhost/api",
    ]

    df = daft.from_pydict({"urls": urls})

    result = df.select(col("urls").url_parse().alias("parsed")).select(col("parsed").struct.get("*")).collect()

    data = result.to_pydict()

    assert data["scheme"][0] == "https"
    assert data["username"][0] == "user"
    assert data["password"][0] == "pass"
    assert data["host"][0] == "example.com"
    assert data["port"][0] == 8080
    assert data["path"][0] == "/path"
    assert data["query"][0] == "query=value"
    assert data["fragment"][0] == "fragment"

    assert data["scheme"][1] == "http"
    assert data["username"][1] is None
    assert data["password"][1] is None
    assert data["host"][1] == "localhost"
    assert data["port"][1] is None
    assert data["path"][1] == "/api"
    assert data["query"][1] is None
    assert data["fragment"][1] is None


def test_url_parse_individual_fields():
    urls = ["https://example.com:443/path?query=value#fragment"]

    df = daft.from_pydict({"urls": urls})

    result = (
        df.select(col("urls").url_parse())
        .select(
            col("urls").struct.get("scheme").alias("scheme"),
            col("urls").struct.get("host").alias("host"),
            col("urls").struct.get("port").alias("port"),
            col("urls").struct.get("path").alias("path"),
            col("urls").struct.get("query").alias("query"),
            col("urls").struct.get("fragment").alias("fragment"),
            col("urls").struct.get("username").alias("username"),
            col("urls").struct.get("password").alias("password"),
        )
        .collect()
    )

    data = result.to_pydict()

    assert data["scheme"][0] == "https"
    assert data["host"][0] == "example.com"
    assert data["port"][0] is None
    assert data["path"][0] == "/path"
    assert data["query"][0] == "query=value"
    assert data["fragment"][0] == "fragment"
    assert data["username"][0] is None
    assert data["password"][0] is None


def test_url_parse_null_handling():
    urls = [
        "https://example.com",
        None,
        "http://localhost",
    ]

    df = daft.from_pydict({"urls": urls})

    result = df.select(col("urls").url_parse().alias("parsed")).select(col("parsed").struct.get("*")).collect()

    data = result.to_pydict()

    assert data["scheme"][0] == "https"
    assert data["host"][0] == "example.com"

    assert data["scheme"][1] is None
    assert data["host"][1] is None
    assert data["path"][1] is None

    assert data["scheme"][2] == "http"
    assert data["host"][2] == "localhost"


def test_url_parse_invalid_urls():
    urls = [
        "https://example.com",
        "invalid-url",
        "",
        "not a url at all",
    ]

    df = daft.from_pydict({"urls": urls})

    result = df.select(col("urls").url_parse().alias("parsed")).select(col("parsed").struct.get("*")).collect()

    data = result.to_pydict()

    assert data["scheme"][0] == "https"
    assert data["host"][0] == "example.com"

    for i in [1, 2, 3]:
        assert data["scheme"][i] is None
        assert data["host"][i] is None
        assert data["path"][i] is None
        assert data["query"][i] is None
        assert data["fragment"][i] is None


def test_url_parse_edge_cases():
    urls = [
        "https://example.com",
        "https://example.com/",
        "https://example.com:443",
        "http://example.com:80",
        "https://example.com?query=value",
        "https://example.com#fragment",
        "mailto:user@example.com",
        "file:///path/to/file",
    ]

    df = daft.from_pydict({"urls": urls})

    result = df.select(col("urls").url_parse()).select(col("urls").struct.get("*")).collect()

    data = result.to_pydict()

    expected_schemes = ["https", "https", "https", "http", "https", "https", "mailto", "file"]
    assert data["scheme"] == expected_schemes

    expected_hosts = [
        "example.com",
        "example.com",
        "example.com",
        "example.com",
        "example.com",
        "example.com",
        None,
        None,
    ]
    assert data["host"] == expected_hosts

    expected_ports = [None, None, None, None, None, None, None, None]
    assert data["port"] == expected_ports


def test_url_parse_github_issue_example():
    urls = [
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.ar/train-00004-of-00007.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.ar/train-00005-of-00007.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.ar/train-00006-of-00007.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.arc/train-00000-of-00001.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.ary/train-00000-of-00001.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.de/train-00013-of-00020.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.de/train-00014-of-00020.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.de/train-00015-of-00020.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.de/train-00016-of-00020.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.de/train-00017-of-00020.parquet",
        "https://huggingface.co/datasets/wikimedia/wikipedia/resolve/main/20231101.de/train-00018-of-00020.parquet",
    ]

    df = daft.from_pydict({"urls": urls})

    result = df.select(col("urls").url_parse()).select(col("urls").struct.get("*")).collect()

    data = result.to_pydict()

    assert all(scheme == "https" for scheme in data["scheme"])
    assert all(host == "huggingface.co" for host in data["host"])

    for path in data["path"]:
        assert path.startswith("/datasets/wikimedia/wikipedia/resolve/main/")
        assert path.endswith(".parquet")

    assert all(query is None for query in data["query"])
    assert all(fragment is None for fragment in data["fragment"])
    assert all(username is None for username in data["username"])
    assert all(password is None for password in data["password"])

    assert all(port is None for port in data["port"])


def test_url_parse_with_authentication():
    urls = [
        "https://user@example.com/path",
        "https://user:pass@example.com/path",
        "ftp://anonymous:email@ftp.example.com/file",
    ]

    df = daft.from_pydict({"urls": urls})

    result = df.select(col("urls").url_parse().alias("parsed")).select(col("parsed").struct.get("*")).collect()

    data = result.to_pydict()

    assert data["username"][0] == "user"
    assert data["password"][0] is None

    assert data["username"][1] == "user"
    assert data["password"][1] == "pass"

    assert data["username"][2] == "anonymous"
    assert data["password"][2] == "email"



from __future__ import annotations

import pytest

import daft


@pytest.mark.integration()
def test_url_download_http(mock_http_image_urls, image_data):
    data = {"urls": mock_http_image_urls}
    df = daft.from_pydict(data)
    df = df.with_column("data", df["urls"].url.download())
    assert df.to_pydict() == {**data, "data": [image_data for _ in range(len(mock_http_image_urls))]}


@pytest.mark.integration()
@pytest.mark.parametrize("status_code", [400, 401, 403, 404, 429, 500, 503])
def test_url_download_http_error_codes(nginx_config, status_code):
    server_url, _ = nginx_config
    data = {"urls": [f"{server_url}/{status_code}.html"]}
    df = daft.from_pydict(data)
    df = df.with_column("data", df["urls"].url.download(on_error="raise"))

    # 404 should always be corner-cased to return FileNotFoundError
    if status_code == 404:
        with pytest.raises(FileNotFoundError):
            df.collect()
    else:
        # NOTE: We may want to add better errors in the future to provide a better
        # user-facing I/O error with the error code
        with pytest.raises(ValueError, match=f"{status_code}"):
            df.collect()



from __future__ import annotations

import copy
from typing import Any

import pytest

TEST_ITEMS = [
    {"sepal_length": 5.1, "sepal_width": 3.5, "petal_length": 1.4, "petal_width": 0.2, "variety": "Setosa"},
    {"sepal_length": 4.9, "sepal_width": 3.0, "petal_length": 1.4, "petal_width": 0.2, "variety": "Setosa"},
    {"sepal_length": 4.7, "sepal_width": 3.2, "petal_length": 1.3, "petal_width": 0.2, "variety": "Setosa"},
]


@pytest.fixture(scope="function")
def valid_data() -> list[dict[str, Any]]:
    return TEST_ITEMS


@pytest.fixture(scope="function")
def missing_value_data() -> list[dict[str, Any]]:
    items = copy.deepcopy(TEST_ITEMS)
    items[0]["sepal_length"] = None
    items[1]["sepal_width"] = float("NaN")
    return items



from __future__ import annotations


def test_with_columns(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    expanded_df = df.with_columns({"foo": df["sepal_width"], "bar": df["sepal_width"] + df["petal_length"]})
    data = expanded_df.to_pydict()
    assert expanded_df.column_names == df.column_names + ["foo", "bar"]
    assert data["foo"] == data["sepal_width"]
    assert data["bar"] == [sw + pl for sw, pl in zip(data["sepal_width"], data["petal_length"])]


def test_with_columns_same_name(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    expanded_df = df.with_columns(
        {"sepal_length": df["sepal_length"] + df["sepal_width"], "petal_length": df["petal_length"] + df["petal_width"]}
    )
    data = expanded_df.to_pydict()
    assert expanded_df.column_names == df.column_names
    expected_sepal_length = [
        valid_data[i]["sepal_length"] + valid_data[i]["sepal_width"] for i in range(len(valid_data))
    ]
    expected_petal_length = [
        valid_data[i]["petal_length"] + valid_data[i]["petal_width"] for i in range(len(valid_data))
    ]
    assert data["sepal_length"] == expected_sepal_length
    assert data["petal_length"] == expected_petal_length


def test_with_columns_empty(make_df, valid_data: list[dict[str, float]]) -> None:
    df = make_df(valid_data)
    expanded_df = df.with_columns({})
    assert expanded_df.column_names == df.column_names
    assert expanded_df.to_pydict() == df.to_pydict()



from __future__ import annotations

import daft


def test_list_map():
    items = [["a", "b", "a"], ["b", "c", "b", "c"]]
    df = daft.from_pydict({"letters": items})
    expected = [["A", "B", "A"], ["B", "C", "B", "C"]]
    actual = df.select(daft.col("letters").list.map(daft.element().str.upper())).to_pydict()["letters"]
    assert actual == expected


def test_list_map_sql():
    items = [["a", "b", "a"], ["b", "c", "b", "c"]]
    df = daft.from_pydict({"letters": items})
    expected = [["A", "B", "A"], ["B", "C", "B", "C"]]
    actual = daft.sql("select list_map(letters, upper(element())) from df", df=df).to_pydict()["letters"]
    assert actual == expected


def test_list_map_with_udf():
    items = [["a", "b", "a"], ["b", "c", "b", "c"]]
    df = daft.from_pydict({"letters": items})
    expected = [["A", "B", "A"], ["B", "C", "B", "C"]]
    actual = df.select(
        daft.col("letters").list.map(daft.element().apply(lambda s: s.upper(), return_dtype=str))
    ).to_pydict()["letters"]
    assert actual == expected


def test_map_chaining():
    df = daft.from_pydict({"numbers": [[1, 2, 3], [4, 5, 6, 7]]})
    actual = df.select(daft.col("numbers").list.map(daft.element() * 2).list.map(daft.element() * 4)).to_pydict()[
        "numbers"
    ]

    expected = [[8, 16, 24], [32, 40, 48, 56]]

    assert actual == expected


def test_map_nested():
    df = daft.from_pydict({"sentences": [["this is a test", "another test"]]})

    words = (
        df.select(daft.col("sentences").list.map(daft.element().str.split(" ")).alias("words"))
        .explode("words")
        .explode("words")
        .to_pydict()["words"]
    )

    expected = ["this", "is", "a", "test", "another", "test"]

    assert words == expected



from __future__ import annotations

import random
from uuid import uuid4

import numpy as np
import pytest

import daft
from daft import DataFrame, col

NUM_SAMPLES = 1_000_000

JOIN_TYPES = ["inner", "left", "right", "outer"]


@pytest.mark.benchmark(group="joins")
@pytest.mark.parametrize(
    "num_samples, num_partitions",
    [(10_000, 1), (10_000, 100)],
    ids=["10_000/1", "10_000/100"],
)
@pytest.mark.parametrize("join_type", JOIN_TYPES)
def test_join_simple(benchmark, num_samples, num_partitions, join_type) -> None:
    """Test simple join performance.

    Keys are consecutive integers; no data payload; one-to-one matches.
    """
    left_arr = np.arange(num_samples)
    np.random.shuffle(left_arr)
    right_arr = np.arange(num_samples)
    np.random.shuffle(right_arr)

    left_table = (
        daft.from_pydict(
            {
                "mycol": left_arr,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )
    right_table = (
        daft.from_pydict(
            {
                "mycol": right_arr,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )

    # Run the benchmark.
    # Run the benchmark.
    def bench_join() -> DataFrame:
        return left_table.join(right_table, on=["mycol"], how=join_type).collect()

    result = benchmark(bench_join)

    # Make sure the result is correct.
    assert (result.sort("mycol").to_pandas()["mycol"].to_numpy() == np.arange(num_samples)).all()


@pytest.mark.benchmark(group="joins")
@pytest.mark.parametrize(
    "num_samples, num_partitions",
    [(10_000, 1), (10_000, 100)],
    ids=["10_000/1", "10_000/100"],
)
@pytest.mark.parametrize("join_type", JOIN_TYPES)
def test_join_largekey(benchmark, num_samples, num_partitions, join_type) -> None:
    """Test the impact of string keys vs integer keys."""
    keys = [str(uuid4()) for _ in range(num_samples)]

    left_keys = keys.copy()
    random.shuffle(left_keys)
    right_keys = keys.copy()
    random.shuffle(right_keys)

    left_table = (
        daft.from_pydict(
            {
                "mycol": left_keys,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )
    right_table = (
        daft.from_pydict(
            {
                "mycol": right_keys,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )

    # Run the benchmark.
    def bench_join() -> DataFrame:
        return left_table.join(right_table, on=["mycol"], how=join_type).collect()

    result = benchmark(bench_join)

    # Make sure the result is correct.
    result_keys = result.to_pydict()["mycol"]
    result_keys.sort()
    keys.sort()
    assert result_keys == keys


@pytest.mark.benchmark(group="joins")
@pytest.mark.parametrize(
    "num_samples, num_partitions",
    [(10_000, 1), (10_000, 100)],
    ids=["10_000/1", "10_000/100"],
)
@pytest.mark.parametrize("join_type", JOIN_TYPES)
def test_join_withdata(benchmark, num_samples, num_partitions, join_type) -> None:
    """Test the impact of data payloads."""
    left_arr = np.arange(num_samples)
    np.random.shuffle(left_arr)
    right_arr = np.arange(num_samples)
    np.random.shuffle(right_arr)

    long_A = "A" * 1024
    long_B = "B" * 1024

    left_table = (
        daft.from_pydict(
            {
                "mykey": left_arr,
                "left_data": [long_A for _ in range(num_samples)],
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )
    right_table = (
        daft.from_pydict(
            {
                "mykey": right_arr,
                "right_data": [long_B for _ in range(num_samples)],
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )

    # Run the benchmark.
    # Run the benchmark.
    def bench_join() -> DataFrame:
        return left_table.join(right_table, on=["mykey"], how=join_type).collect()

    result = benchmark(bench_join)

    # Make sure the result is correct.
    assert (result.sort("mykey").to_pandas()["mykey"].to_numpy() == np.arange(num_samples)).all()
    assert result.groupby("left_data", "right_data").agg(col("mykey").count()).to_pydict() == {
        "left_data": [long_A],
        "right_data": [long_B],
        "mykey": [num_samples],
    }


@pytest.mark.benchmark(group="joins")
@pytest.mark.parametrize(
    "left_bigger",
    [True, False],
    ids=["left_bigger", "right_bigger"],
)
@pytest.mark.parametrize("num_partitions", [1, 10], ids=["1part", "10part"])
@pytest.mark.parametrize("join_type", JOIN_TYPES)
def test_broadcast_join(benchmark, left_bigger, num_partitions, join_type) -> None:
    """Test the performance of joining a smaller table to a bigger table.

    The cardinality is one-to-many.
    """
    small_length = 1_000
    big_factor = 10

    small_arr = np.arange(small_length)
    np.random.shuffle(small_arr)

    big_arr = np.concatenate([np.arange(small_length) for _ in range(big_factor)])
    np.random.shuffle(big_arr)

    small_table = daft.from_pydict(
        {
            "keys": small_arr,
            "data": [str(x) for x in small_arr],
        }
    ).collect()
    big_table = (
        daft.from_pydict(
            {
                "keys": big_arr,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )

    # Run the benchmark.
    def bench_join() -> DataFrame:
        if left_bigger:
            return big_table.join(small_table, on=["keys"], how=join_type).collect()
        else:
            return small_table.join(big_table, on=["keys"], how=join_type).collect()

    result = benchmark(bench_join)

    # Make sure the result is correct.
    data = result.sort("keys").to_pydict()["data"]
    assert data[:big_factor] == ["0"] * big_factor
    assert data[-big_factor:] == [str(small_length - 1)] * big_factor


@pytest.mark.benchmark(group="joins")
@pytest.mark.parametrize(
    "num_samples, num_partitions",
    [(10_000, 1), (10_000, 100)],
    ids=["10_000/1", "10_000/100"],
)
@pytest.mark.parametrize("num_columns", [1, 4])
@pytest.mark.parametrize("join_type", JOIN_TYPES)
def test_multicolumn_joins(benchmark, num_columns, num_samples, num_partitions, join_type) -> None:
    """Evaluate the performance impact of using additional columns in the join.

    The join cardinality is the same for all cases;
    redundant columns are used for the multicolumn joins.
    """
    left_arr = np.arange(num_samples)
    np.random.shuffle(left_arr)
    right_arr = np.arange(num_samples)
    np.random.shuffle(right_arr)

    left_table = (
        daft.from_pydict(
            {
                "nums_a": left_arr * 17 % 9,
                "nums_b": left_arr * 17 % 10,
                "nums_c": left_arr * 17 % 11,
                "nums": left_arr,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )
    right_table = (
        daft.from_pydict(
            {
                "nums_a": right_arr * 17 % 9,
                "nums_b": right_arr * 17 % 10,
                "nums_c": right_arr * 17 % 11,
                "nums": right_arr,
            }
        )
        .into_partitions(num_partitions)
        .collect()
    )

    # Run the benchmark.
    # Run the benchmark.
    def bench_join() -> DataFrame:
        # Use the unique column "nums" plus some redundant columns.
        join_on = ["nums_a", "nums_b", "nums_c", "nums"][-num_columns:]
        return left_table.join(right_table, on=join_on, how=join_type).collect()

    result = benchmark(bench_join)

    # Make sure the result is correct.
    assert (result.sort("nums").to_pandas()["nums"].to_numpy() == np.arange(num_samples)).all()



from __future__ import annotations

from daft.daft import IOConfig, S3Config
from daft.io.object_store_options import io_config_to_storage_options


def test_convert_to_s3_config():
    s3_config = S3Config(
        region_name="us-east-2",
        endpoint_url="https://s3.us-east-2.amazonaws.com",
        key_id="dummy_ak",
        access_key="dummy_sk",
        use_ssl=True,
        verify_ssl=True,
        connect_timeout_ms=1000,
        anonymous=True,
        force_virtual_addressing=True,
    )
    expected_s3_config = {
        "region": "us-east-2",
        "endpoint_url": "https://dummy_bucket.s3.us-east-2.amazonaws.com",
        "access_key_id": "dummy_ak",
        "secret_access_key": "dummy_sk",
        "allow_http": "false",
        "allow_invalid_certificates": "false",
        "connect_timeout": "1000ms",
        "skip_signature": "true",
        "virtual_hosted_style_request": "true",
    }
    table_uri = "s3://dummy_bucket/path"

    # convert when endpoint doesn't contain bucket for virtual host style
    config = io_config_to_storage_options(IOConfig(s3=s3_config), table_uri=table_uri)
    assert config == expected_s3_config

    # convert when endpoint contains bucket for virtual host style
    new_s3_config = s3_config.replace(endpoint_url="https://dummy_bucket.s3.us-east-2.amazonaws.com")
    config = io_config_to_storage_options(IOConfig(s3=new_s3_config), table_uri=table_uri)
    assert config == expected_s3_config

    # convert when force virtual addressing is false
    new_s3_config = s3_config.replace(force_virtual_addressing=False)
    config = io_config_to_storage_options(IOConfig(s3=new_s3_config), table_uri=table_uri)
    expected_s3_config.update(endpoint_url="https://s3.us-east-2.amazonaws.com")
    expected_s3_config.pop("virtual_hosted_style_request")
    assert config == expected_s3_config



from __future__ import annotations

import lance
import pyarrow as pa
import pytest

import daft

TABLE_NAME = "my_table"
data = {
    "vector": [[1.1, 1.2], [0.2, 1.8]],
    "lat": [45.5, 40.1],
    "long": [-122.7, -74.1],
}

PYARROW_LOWER_BOUND_SKIP = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) < (9, 0, 0)
pytestmark = pytest.mark.skipif(PYARROW_LOWER_BOUND_SKIP, reason="lance not supported on old versions of pyarrow")


@pytest.fixture(scope="function")
def lance_dataset_path(tmp_path_factory):
    tmp_dir = tmp_path_factory.mktemp("lance")
    lance.write_dataset(pa.Table.from_pydict(data), tmp_dir)
    yield str(tmp_dir)


def test_lancedb_read(lance_dataset_path):
    df = daft.read_lance(lance_dataset_path)
    assert df.to_pydict() == data


def test_lancedb_read_column_selection(lance_dataset_path):
    df = daft.read_lance(lance_dataset_path)
    df = df.select("vector")
    assert df.to_pydict() == {"vector": data["vector"]}


def test_lancedb_read_filter(lance_dataset_path):
    df = daft.read_lance(lance_dataset_path)
    df = df.where(df["lat"] > 45)
    df = df.select("vector")
    assert df.to_pydict() == {"vector": data["vector"][:1]}


def test_lancedb_read_limit(lance_dataset_path):
    df = daft.read_lance(lance_dataset_path)
    df = df.limit(1)
    df = df.select("vector")
    assert df.to_pydict() == {"vector": data["vector"][:1]}


def test_lancedb_with_version(lance_dataset_path):
    df = daft.read_lance(lance_dataset_path, version=1)
    assert df.to_pydict() == data



from __future__ import annotations

import pytest

from daft.expressions import col
from tests.expressions.typing.conftest import (
    assert_typing_resolve_vs_runtime_behavior,
    is_comparable,
    is_numeric,
)


@pytest.mark.parametrize(
    "op",
    [
        pytest.param(lambda x: x.min(), id="min"),
        pytest.param(lambda x: x.max(), id="max"),
    ],
)
def test_comparable_aggs(unary_data_fixture, op):
    arg = unary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=(unary_data_fixture,),
        expr=op(col(arg.name())),
        run_kernel=lambda: op(arg),
        resolvable=is_comparable(arg.datatype()),
    )


@pytest.mark.parametrize(
    "op",
    [
        pytest.param(lambda x: x.sum(), id="sum"),
        pytest.param(lambda x: x.mean(), id="mean"),
    ],
)
def test_numeric_aggs(unary_data_fixture, op):
    arg = unary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=(unary_data_fixture,),
        expr=op(col(arg.name())),
        run_kernel=lambda: op(arg),
        resolvable=is_numeric(arg.datatype()),
    )


def test_decimal_sum(decimal_unary_data_fixture):
    """A copy of the above but for decimal types that do not more widely support
    numeric operations. When they do and can be added to ALL_DTYPES and resolve
    is_numeric to True, this test can be removed.
    """  # noqa: D205
    arg = decimal_unary_data_fixture

    def op(x):
        return x.sum()

    assert_typing_resolve_vs_runtime_behavior(
        data=(decimal_unary_data_fixture,),
        expr=op(col(arg.name())),
        run_kernel=lambda: op(arg),
        resolvable=True,
    )


def test_count(unary_data_fixture):
    arg = unary_data_fixture
    assert_typing_resolve_vs_runtime_behavior(
        data=(unary_data_fixture,),
        expr=col(arg.name()).count(),
        run_kernel=lambda: arg.count(),
        resolvable=True,
    )



from __future__ import annotations

import random

import pandas as pd
import pytest

from daft import Window, col
from daft.context import get_context
from daft.functions import dense_rank, rank
from tests.conftest import assert_df_equals, get_tests_daft_runner_name

pytestmark = pytest.mark.skipif(
    get_tests_daft_runner_name() == "ray"
    and get_context().daft_execution_config.use_experimental_distributed_engine is False,
    reason="requires Native Runner or Flotilla to be in use",
)


def test_basic_lag_function(make_df):
    """Test basic lag function with default offset."""
    random.seed(42)

    data = []
    expected_data = []

    for category in ["A", "B", "C"]:
        values = random.sample(range(10, 100), 10)
        for ts, value in enumerate(values):
            data.append({"category": category, "ts": ts, "value": value})
            expected_data.append(
                {"category": category, "ts": ts, "value": value, "lagged_value": None if ts == 0 else values[ts - 1]}
            )

    df = make_df(data)

    window_spec = Window().partition_by("category").order_by("ts", desc=False)

    result = df.select(
        col("category"), col("ts"), col("value"), col("value").lag(1).over(window_spec).alias("lagged_value")
    ).collect()

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected_data), sort_key=["category", "ts"], check_dtype=False)


def test_basic_lead_function(make_df):
    """Test basic lead function with default offset."""
    random.seed(43)

    data = []
    expected_data = []

    for category in ["A", "B", "C"]:
        values = random.sample(range(10, 100), 10)
        for ts, value in enumerate(values):
            data.append({"category": category, "ts": ts, "value": value})
            expected_data.append(
                {"category": category, "ts": ts, "value": value, "lead_value": None if ts == 9 else values[ts + 1]}
            )

    df = make_df(data)

    window_spec = Window().partition_by("category").order_by("ts", desc=False)

    result = df.select(
        col("category"), col("ts"), col("value"), col("value").lead(1).over(window_spec).alias("lead_value")
    ).collect()

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected_data), sort_key=["category", "ts"], check_dtype=False)


@pytest.mark.parametrize(
    "desc,nulls_first,expected_lead",
    [
        (False, False, [2, 3, 4, None, 6, 7, 8, None, 10, 11, 12, None]),
        (False, True, [2, 3, None, 1, 6, 7, None, 5, 10, 11, None, 9]),
        (True, False, [4, 1, 2, None, 8, 5, 6, None, 12, 9, 10, None]),
        (True, True, [None, 1, 2, 3, None, 5, 6, 7, None, 9, 10, 11]),
    ],
)
def test_lead_nulls_first_or_last(make_df, desc, nulls_first, expected_lead):
    """Test lag and lead functions with nulls first or last."""
    data = [
        {"id": 1, "category": "A", "ts": 1},
        {"id": 2, "category": "A", "ts": 2},
        {"id": 3, "category": "A", "ts": 3},
        {"id": 4, "category": "A", "ts": None},
        {"id": 5, "category": "B", "ts": 1},
        {"id": 6, "category": "B", "ts": 2},
        {"id": 7, "category": "B", "ts": 3},
        {"id": 8, "category": "B", "ts": None},
        {"id": 9, "category": "C", "ts": 1},
        {"id": 10, "category": "C", "ts": 2},
        {"id": 11, "category": "C", "ts": 3},
        {"id": 12, "category": "C", "ts": None},
    ]

    df = make_df(data)
    window_spec = Window().partition_by("category").order_by("ts", desc=desc, nulls_first=nulls_first)
    result = df.select(
        col("id"), col("category"), col("ts"), col("id").lead(1).over(window_spec).alias("lead_value")
    ).collect()

    expected = {
        "id": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        "category": ["A", "A", "A", "A", "B", "B", "B", "B", "C", "C", "C", "C"],
        "ts": [1, 2, 3, None, 1, 2, 3, None, 1, 2, 3, None],
        "lead_value": expected_lead,
    }

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected), sort_key=["id"], check_dtype=False)


def test_lag_lead_with_different_offsets(make_df):
    """Test lag and lead functions with different offset values."""
    random.seed(44)

    data = []
    expected_data = []

    for category in ["A", "B"]:
        values = random.sample(range(10, 100), 10)
        for ts in range(7):
            value = values[ts]
            data.append({"category": category, "ts": ts, "value": value})

            expected_row = {
                "category": category,
                "ts": ts,
                "value": value,
                "lag_1": None if ts < 1 else values[ts - 1],
                "lag_2": None if ts < 2 else values[ts - 2],
                "lag_3": None if ts < 3 else values[ts - 3],
                "lead_1": None if ts >= 6 else values[ts + 1],
                "lead_2": None if ts >= 5 else values[ts + 2],
                "lead_3": None if ts >= 4 else values[ts + 3],
            }

            expected_data.append(expected_row)

    df = make_df(data)

    window_spec = Window().partition_by("category").order_by("ts", desc=False)

    result = df.select(
        col("category"),
        col("ts"),
        col("value"),
        col("value").lag(1).over(window_spec).alias("lag_1"),
        col("value").lag(2).over(window_spec).alias("lag_2"),
        col("value").lag(3).over(window_spec).alias("lag_3"),
        col("value").lead(1).over(window_spec).alias("lead_1"),
        col("value").lead(2).over(window_spec).alias("lead_2"),
        col("value").lead(3).over(window_spec).alias("lead_3"),
    ).collect()

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected_data), sort_key=["category", "ts"], check_dtype=False)


def test_lag_lead_with_col_default(make_df):
    """Test lag and lead functions with column default values."""
    random.seed(45)

    data = []
    expected_data = []

    for category in ["A", "B"]:
        defaults = random.sample(range(10, 100), 50)
        values = random.sample(range(10, 100), 50)
        for ts in range(50):
            value = values[ts]
            data.append({"category": category, "ts": ts, "default": defaults[ts], "value": value})

            expected_row = {
                "category": category,
                "ts": ts,
                "default": defaults[ts],
                "value": value,
                "lag_10": 2 * defaults[ts] if ts < 10 else values[ts - 10],
                "lead_10": 2 * defaults[ts] if ts >= 40 else values[ts + 10],
                "lag_50": 2 * defaults[ts],
                "lead_50": 2 * defaults[ts],
                "lag_15_default_99": 99 if ts < 15 else values[ts - 15],
                "lead_15_default_99": 99 if ts >= 35 else values[ts + 15],
            }
            expected_data.append(expected_row)

    df = make_df(data)

    window_spec = Window().partition_by("category").order_by("ts", desc=False)

    result = df.select(
        col("category"),
        col("ts"),
        col("default"),
        col("value"),
        col("value").lag(10, default=2 * col("default")).over(window_spec).alias("lag_10"),
        col("value").lead(10, default=2 * col("default")).over(window_spec).alias("lead_10"),
        col("value").lag(50, default=2 * col("default")).over(window_spec).alias("lag_50"),
        col("value").lead(50, default=2 * col("default")).over(window_spec).alias("lead_50"),
        col("value").lag(15, default=99).over(window_spec).alias("lag_15_default_99"),
        col("value").lead(15, default=99).over(window_spec).alias("lead_15_default_99"),
    ).collect()

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected_data), sort_key=["category", "ts"], check_dtype=False)


def test_lag_lead_with_multiple_partitions(make_df):
    """Test lag and lead functions with multiple partition columns."""
    random.seed(45)

    data = []
    expected_data = []

    for category in ["A", "B", "C"]:
        for subcategory in [1, 2, 3]:
            values = random.sample(range(10, 100), 4)
            for ts in range(4):
                value = values[ts]
                data.append({"category": category, "subcategory": subcategory, "ts": ts, "value": value})

                lag_value = None if ts == 0 else values[ts - 1]
                lead_value = None if ts == 3 else values[ts + 1]

                expected_data.append(
                    {
                        "category": category,
                        "subcategory": subcategory,
                        "ts": ts,
                        "value": value,
                        "lag_value": lag_value,
                        "lead_value": lead_value,
                    }
                )

    df = make_df(data)

    combined_window = Window().partition_by(["category", "subcategory"]).order_by("ts", desc=False)

    result = df.select(
        col("category"),
        col("subcategory"),
        col("ts"),
        col("value"),
        col("value").lag(1).over(combined_window).alias("lag_value"),
        col("value").lead(1).over(combined_window).alias("lead_value"),
    ).collect()

    assert_df_equals(
        result.to_pandas(), pd.DataFrame(expected_data), sort_key=["category", "subcategory", "ts"], check_dtype=False
    )


def test_lag_lead_for_delta_calculation(make_df):
    """Test using lag and lead functions to calculate deltas (differences between values)."""
    random.seed(46)

    data = []
    expected_data = []

    for category in ["A", "B"]:
        values = random.sample(range(10, 100), 5)
        for ts in range(5):
            value = values[ts]
            data.append({"category": category, "ts": ts, "value": value})

            delta_from_prev = None if ts == 0 else value - values[ts - 1]
            delta_to_next = None if ts == 4 else values[ts + 1] - value
            delta_from_2_prev = None if ts < 2 else value - values[ts - 2]

            expected_data.append(
                {
                    "category": category,
                    "ts": ts,
                    "value": value,
                    "delta_from_prev": delta_from_prev,
                    "delta_to_next": delta_to_next,
                    "delta_from_2_prev": delta_from_2_prev,
                }
            )

    df = make_df(data)

    window_spec = Window().partition_by("category").order_by("ts", desc=False)

    result = df.select(
        col("category"),
        col("ts"),
        col("value"),
        (col("value") - col("value").lag(1).over(window_spec)).alias("delta_from_prev"),
        (col("value").lead(1).over(window_spec) - col("value")).alias("delta_to_next"),
        (col("value") - col("value").lag(2).over(window_spec)).alias("delta_from_2_prev"),
    ).collect()

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected_data), sort_key=["category", "ts"], check_dtype=False)


def test_lag_lead_with_zero_offset(make_df):
    """Test lag and lead functions with zero offset (should return current row)."""
    random.seed(47)

    data = []
    expected_data = []

    for category in ["A", "B", "C"]:
        values = random.sample(range(10, 100), 3)
        for ts in range(3):
            value = values[ts]
            data.append({"category": category, "ts": ts, "value": value})

            expected_data.append({"category": category, "ts": ts, "value": value, "lag_0": value, "lead_0": value})

    df = make_df(data)

    window_spec = Window().partition_by("category").order_by("ts", desc=False)

    result = df.select(
        col("category"),
        col("ts"),
        col("value"),
        col("value").lag(0).over(window_spec).alias("lag_0"),
        col("value").lead(0).over(window_spec).alias("lead_0"),
    ).collect()

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected_data), sort_key=["category", "ts"], check_dtype=False)


def test_lag_lead_with_large_offset(make_df):
    """Test lag and lead functions with offset larger than partition size."""
    random.seed(48)

    data = []
    expected_data = []

    for category in ["A", "B", "C"]:
        partition_size = 4
        values = random.sample(range(10, 100), partition_size)

        for ts in range(partition_size):
            value = values[ts]
            data.append({"category": category, "ts": ts, "value": value})

            large_offset = 10
            expected_data.append(
                {
                    "category": category,
                    "ts": ts,
                    "value": value,
                    f"lag_{large_offset}": None,
                    f"lead_{large_offset}": None,
                }
            )

    df = make_df(data)

    large_offset = 10
    window_spec = Window().partition_by("category").order_by("ts", desc=False)

    result = df.select(
        col("category"),
        col("ts"),
        col("value"),
        col("value").lag(large_offset).over(window_spec).alias(f"lag_{large_offset}"),
        col("value").lead(large_offset).over(window_spec).alias(f"lead_{large_offset}"),
    ).collect()

    result = result.to_pandas().map(lambda x: None if pd.isna(x) else x)

    assert_df_equals(result, pd.DataFrame(expected_data), sort_key=["category", "ts"], check_dtype=False)


def test_lag_lead_with_default_values(make_df):
    """Test lag and lead functions with default values for NULL results."""
    random.seed(49)

    data = []
    expected_data = []
    default_value = -999

    for category in ["A", "B"]:
        values = random.sample(range(10, 100), 5)
        for ts in range(5):
            value = values[ts]
            data.append({"category": category, "ts": ts, "value": value})

            lag_with_default = default_value if ts == 0 else values[ts - 1]
            lead_with_default = default_value if ts == 4 else values[ts + 1]

            lag_large_with_default = default_value
            lead_large_with_default = default_value

            expected_data.append(
                {
                    "category": category,
                    "ts": ts,
                    "value": value,
                    "lag_with_default": lag_with_default,
                    "lead_with_default": lead_with_default,
                    "lag_large_with_default": lag_large_with_default,
                    "lead_large_with_default": lead_large_with_default,
                }
            )

    df = make_df(data)

    window_spec = Window().partition_by("category").order_by("ts", desc=False)

    result = df.select(
        col("category"),
        col("ts"),
        col("value"),
        col("value").lag(1, default=default_value).over(window_spec).alias("lag_with_default"),
        col("value").lead(1, default=default_value).over(window_spec).alias("lead_with_default"),
        col("value").lag(10, default=default_value).over(window_spec).alias("lag_large_with_default"),
        col("value").lead(10, default=default_value).over(window_spec).alias("lead_large_with_default"),
    ).collect()

    assert_df_equals(result.to_pandas(), pd.DataFrame(expected_data), sort_key=["category", "ts"], check_dtype=False)


def test_all_partition_order_by(make_df):
    """Test combining multiple window functions across different partition schemes."""
    random.seed(42)

    data = []

    for category in ["A", "B", "C"]:
        for subcategory in [1, 2, 3]:
            for ts in range(5):
                value = random.randint(10, 100)
                data.append({"category": category, "subcategory": subcategory, "ts": ts, "value": value})

    df = make_df(data)

    category_value_window = Window().partition_by("category").order_by("value", desc=False)
    subcategory_value_window = Window().partition_by("subcategory").order_by("value", desc=False)
    combined_value_window = Window().partition_by(["category", "subcategory"]).order_by("value", desc=False)
    combined_window = Window().partition_by(["category", "subcategory"]).order_by("ts", desc=False)

    result = df.select(
        col("category"),
        col("subcategory"),
        col("ts"),
        col("value"),
        rank().over(category_value_window).alias("category_rank"),
        dense_rank().over(category_value_window).alias("category_dense_rank"),
        rank().over(subcategory_value_window).alias("subcategory_rank"),
        dense_rank().over(subcategory_value_window).alias("subcategory_dense_rank"),
        rank().over(combined_value_window).alias("combined_rank"),
        dense_rank().over(combined_value_window).alias("combined_dense_rank"),
        col("value").lag(1).over(combined_window).alias("combined_lag"),
        col("value").lead(1).over(combined_window).alias("combined_lead"),
    ).collect()

    result_dict = result.to_pydict()

    for category in ["A", "B", "C"]:
        category_indices = [i for i, cat in enumerate(result_dict["category"]) if cat == category]
        category_values = [result_dict["value"][i] for i in category_indices]

        sorted_values = sorted(set(category_values))
        value_to_dense_rank = {value: i + 1 for i, value in enumerate(sorted_values)}

        value_counts = {}
        for value in category_values:
            value_counts[value] = value_counts.get(value, 0) + 1

        value_to_rank = {}
        current_rank = 1
        for value in sorted_values:
            value_to_rank[value] = current_rank
            current_rank += value_counts[value]

        for idx in category_indices:
            value = result_dict["value"][idx]
            expected_rank = value_to_rank[value]
            expected_dense_rank = value_to_dense_rank[value]

            assert (
                result_dict["category_rank"][idx] == expected_rank
            ), f"Incorrect rank for value {value} in category {category}"
            assert (
                result_dict["category_dense_rank"][idx] == expected_dense_rank
            ), f"Incorrect dense_rank for value {value} in category {category}"

    for subcategory in [1, 2, 3]:
        subcategory_indices = [i for i, sc in enumerate(result_dict["subcategory"]) if sc == subcategory]
        subcategory_values = [result_dict["value"][i] for i in subcategory_indices]

        sorted_values = sorted(set(subcategory_values))
        value_to_dense_rank = {value: i + 1 for i, value in enumerate(sorted_values)}

        value_counts = {}
        for value in subcategory_values:
            value_counts[value] = value_counts.get(value, 0) + 1

        value_to_rank = {}
        current_rank = 1
        for value in sorted_values:
            value_to_rank[value] = current_rank
            current_rank += value_counts[value]

        for idx in subcategory_indices:
            value = result_dict["value"][idx]
            expected_rank = value_to_rank[value]
            expected_dense_rank = value_to_dense_rank[value]

            assert (
                result_dict["subcategory_rank"][idx] == expected_rank
            ), f"Incorrect rank for value {value} in subcategory {subcategory}"
            assert (
                result_dict["subcategory_dense_rank"][idx] == expected_dense_rank
            ), f"Incorrect dense_rank for value {value} in subcategory {subcategory}"

    for category in ["A", "B", "C"]:
        for subcategory in [1, 2, 3]:
            combined_indices = [
                i
                for i, (cat, sc) in enumerate(zip(result_dict["category"], result_dict["subcategory"]))
                if cat == category and sc == subcategory
            ]
            combined_values = [result_dict["value"][i] for i in combined_indices]

            sorted_values = sorted(set(combined_values))
            value_to_dense_rank = {value: i + 1 for i, value in enumerate(sorted_values)}

            value_counts = {}
            for value in combined_values:
                value_counts[value] = value_counts.get(value, 0) + 1

            value_to_rank = {}
            current_rank = 1
            for value in sorted_values:
                value_to_rank[value] = current_rank
                current_rank += value_counts[value]

            for idx in combined_indices:
                value = result_dict["value"][idx]
                expected_rank = value_to_rank[value]
                expected_dense_rank = value_to_dense_rank[value]

                assert (
                    result_dict["combined_rank"][idx] == expected_rank
                ), f"Incorrect rank for value {value} in {category}-{subcategory}"
                assert (
                    result_dict["combined_dense_rank"][idx] == expected_dense_rank
                ), f"Incorrect dense_rank for value {value} in {category}-{subcategory}"

    for category in ["A", "B", "C"]:
        for subcategory in [1, 2, 3]:
            combined_indices = [
                i
                for i, (cat, sc) in enumerate(zip(result_dict["category"], result_dict["subcategory"]))
                if cat == category and sc == subcategory
            ]

            for ts in range(5):
                current_ts_indices = [i for i in combined_indices if result_dict["ts"][i] == ts]

                if not current_ts_indices:
                    continue

                idx = current_ts_indices[0]

                if ts == 0:
                    assert (
                        result_dict["combined_lag"][idx] is None
                    ), f"First ts in {category}-{subcategory} should have None lag"
                else:
                    prev_ts_indices = [i for i in combined_indices if result_dict["ts"][i] == ts - 1]
                    if prev_ts_indices:
                        prev_idx = prev_ts_indices[0]
                        prev_value = result_dict["value"][prev_idx]
                        assert (
                            result_dict["combined_lag"][idx] == prev_value
                        ), f"Incorrect lag for {category}-{subcategory}, ts {ts}"

                if ts == 4:
                    assert (
                        result_dict["combined_lead"][idx] is None
                    ), f"Last ts in {category}-{subcategory} should have None lead"
                else:
                    next_ts_indices = [i for i in combined_indices if result_dict["ts"][i] == ts + 1]
                    if next_ts_indices:
                        next_idx = next_ts_indices[0]
                        next_value = result_dict["value"][next_idx]
                        assert (
                            result_dict["combined_lead"][idx] == next_value
                        ), f"Incorrect lead for {category}-{subcategory}, ts {ts}"



import os
from datetime import datetime, timezone

import gspread

import daft


def upload_to_google_sheets(worksheet, data):
    gc = gspread.service_account()

    sh = gc.open_by_url(
        "https://docs.google.com/spreadsheets/d/1d6pXsIsBkjjM93GYtoiF83WXvJXR4vFgFQdmG05u8eE/edit?gid=0#gid=0"
    )
    ws = sh.worksheet(worksheet)
    ws.append_row(data)


def get_run_metadata():
    return {
        "started at": datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S.%f"),
        "daft version": daft.__version__,
        "github ref": os.getenv("GITHUB_REF_NAME"),
        "github sha": os.getenv("GITHUB_SHA"),
    }



"""Run TPC-H benchmarks with native runner on local Parquet data and upload results to Google sheets.

Expects tables as Parquet files in "/tmp/tpch-data/"
"""

from __future__ import annotations

import time

import daft
import daft.context
from benchmarking.tpch.answers_sql import get_answer
from tools.ci_bench_utils import get_run_metadata, upload_to_google_sheets


def get_df(name):
    return daft.read_parquet(f"/tmp/tpch-data/{name}/*")


def run_benchmark():
    results = {}

    for q in range(1, 23):
        print(f"Running TPC-H Q{q}... ", end="", flush=True)

        daft_df = get_answer(q, get_df)

        start = time.perf_counter()
        daft_df.collect()
        end = time.perf_counter()

        results[q] = end - start

        print(f"done in {results[q]:.2f}s")

    return results


def main():
    daft.context.set_runner_native()

    metadata = get_run_metadata()

    results = run_benchmark()

    data_dict = {**metadata, **results}

    print("Results:")
    print(data_dict)

    upload_to_google_sheets("Local TPC-H", list(data_dict.values()))


if __name__ == "__main__":
    main()



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations

from typing import Optional, Union

from daft import context
from daft.api_annotations import PublicAPI
from daft.daft import (
    FileFormatConfig,
    IOConfig,
    ParquetSourceConfig,
    StorageConfig,
)
from daft.dataframe import DataFrame
from daft.datatype import DataType, TimeUnit
from daft.io.common import get_tabular_files_scan


@PublicAPI
def read_parquet(
    path: Union[str, list[str]],
    row_groups: Optional[list[list[int]]] = None,
    infer_schema: bool = True,
    schema: Optional[dict[str, DataType]] = None,
    io_config: Optional[IOConfig] = None,
    file_path_column: Optional[str] = None,
    hive_partitioning: bool = False,
    coerce_int96_timestamp_unit: Optional[Union[str, TimeUnit]] = None,
    schema_hints: Optional[dict[str, DataType]] = None,
    _multithreaded_io: Optional[bool] = None,
    _chunk_size: Optional[int] = None,  # A hidden parameter for testing purposes.
) -> DataFrame:
    """Creates a DataFrame from Parquet file(s).

    Args:
        path (str): Path to Parquet file (allows for wildcards)
        row_groups (List[int] or List[List[int]]): List of row groups to read corresponding to each file.
        infer_schema (bool): Whether to infer the schema of the Parquet, defaults to True.
        schema (dict[str, DataType]): A schema that is used as the definitive schema for the Parquet file if infer_schema is False, otherwise it is used as a schema hint that is applied after the schema is inferred.
        io_config (IOConfig): Config to be used with the native downloader
        file_path_column: Include the source path(s) as a column with this name. Defaults to None.
        hive_partitioning: Whether to infer hive_style partitions from file paths and include them as columns in the Dataframe. Defaults to False.
        coerce_int96_timestamp_unit: TimeUnit to coerce Int96 TimeStamps to. e.g.: [ns, us, ms], Defaults to None.
        _multithreaded_io: Whether to use multithreading for IO threads. Setting this to False can be helpful in reducing
            the amount of system resources (number of connections and thread contention) when running in the Ray runner.
            Defaults to None, which will let Daft decide based on the runner it is currently using.

    Returns:
        DataFrame: parsed DataFrame

    Examples:
        >>> df = daft.read_parquet("/path/to/file.parquet")
        >>> df = daft.read_parquet("/path/to/directory")
        >>> df = daft.read_parquet("/path/to/files-*.parquet")
        >>> df = daft.read_parquet("s3://path/to/files-*.parquet")
        >>> df = daft.read_parquet("gs://path/to/files-*.parquet")

    """
    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config

    if isinstance(path, list) and len(path) == 0:
        raise ValueError("Cannot read DataFrame from from empty list of Parquet filepaths")

    if schema_hints is not None:
        raise ValueError(
            "Specifying schema_hints is deprecated from Daft version >= 0.3.0! Instead, please use the 'schema' and 'infer_schema' arguments."
        )

    # If running on Ray, we want to limit the amount of concurrency and requests being made.
    # This is because each Ray worker process receives its own pool of thread workers and connections
    multithreaded_io = (
        (context.get_context().get_or_create_runner().name != "ray") if _multithreaded_io is None else _multithreaded_io
    )

    if isinstance(coerce_int96_timestamp_unit, str):
        coerce_int96_timestamp_unit = TimeUnit.from_str(coerce_int96_timestamp_unit)

    pytimeunit = coerce_int96_timestamp_unit._timeunit if coerce_int96_timestamp_unit is not None else None

    if isinstance(path, list) and row_groups is not None and len(path) != len(row_groups):
        raise ValueError("row_groups must be the same length as the list of paths provided.")
    if isinstance(row_groups, list) and not isinstance(path, list):
        raise ValueError("row_groups are only supported when reading multiple non-globbed/wildcarded files")

    file_format_config = FileFormatConfig.from_parquet_config(
        ParquetSourceConfig(coerce_int96_timestamp_unit=pytimeunit, row_groups=row_groups, chunk_size=_chunk_size)
    )
    storage_config = StorageConfig(multithreaded_io, io_config)

    builder = get_tabular_files_scan(
        path=path,
        infer_schema=infer_schema,
        schema=schema,
        file_format_config=file_format_config,
        storage_config=storage_config,
        file_path_column=file_path_column,
        hive_partitioning=hive_partitioning,
    )
    return DataFrame(builder)



from __future__ import annotations

import pytest

from daft.expressions import col


def test_count_rows(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Count rows for the entire table."""
    daft_df_row_count = daft_df.repartition(repartition_nparts).count_rows()
    assert daft_df_row_count == service_requests_csv_pd_df.shape[0]


def test_dataframe_count_no_args(daft_df, service_requests_csv_pd_df):
    """Counts rows using `df.count()` without any arguments."""
    results = daft_df.count().to_pydict()
    assert "count" in results
    assert len(results["count"]) == 1
    assert results["count"][0] == service_requests_csv_pd_df.shape[0]


def test_filtered_count_rows(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Count rows on a table filtered by a certain condition."""
    daft_df_row_count = daft_df.repartition(repartition_nparts).where(col("Borough") == "BROOKLYN").count_rows()

    pd_df_row_count = len(service_requests_csv_pd_df[service_requests_csv_pd_df["Borough"] == "BROOKLYN"])
    assert daft_df_row_count == pd_df_row_count


@pytest.mark.parametrize(
    "keys",
    [
        pytest.param(["Borough"], id="NumGroupByKeys:1"),
        pytest.param(["Borough", "Complaint Type"], id="NumGroupByKeys:2"),
    ],
)
def test_groupby_count_rows(daft_df, service_requests_csv_pd_df, repartition_nparts, keys, with_morsel_size):
    """Count rows after group by."""
    daft_df = daft_df.repartition(repartition_nparts).groupby(*[col(k) for k in keys]).sum(col("Unique Key"))
    service_requests_csv_pd_df = service_requests_csv_pd_df.groupby(keys).sum("Unique Key").reset_index()
    assert daft_df.count_rows() == len(service_requests_csv_pd_df)


def test_dataframe_length_after_collect(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Count rows after group by."""
    daft_df = daft_df.repartition(repartition_nparts).collect()
    assert len(daft_df) == len(service_requests_csv_pd_df)


def test_dataframe_length_before_collect(daft_df):
    """Count rows for the entire table."""
    with pytest.raises(RuntimeError):
        len(daft_df)


def test_count_rows_after_collect(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Count rows for the entire table."""
    daft_df = daft_df.repartition(repartition_nparts).select(col("Unique Key")).where(col("Unique Key") > 1).collect()
    expected_count = len(daft_df)
    assert daft_df.count_rows() == expected_count



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar, get_type_hints, overload

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

from daft.datatype import DataType
from daft.expressions import Expression
from daft.series import Series

if TYPE_CHECKING:
    from daft.daft import PyDataType, PySeries


P = ParamSpec("P")
T = TypeVar("T")


class RowWiseUdf(Generic[P, T]):
    def __init__(self, fn: Callable[P, T], return_dtype: DataType | None):
        self._inner = fn

        module_name = getattr(fn, "__module__")
        qual_name: str = getattr(fn, "__qualname__")
        if module_name:
            self.name = f"{module_name}.{qual_name}"
        else:
            self.name = qual_name

        type_hints = get_type_hints(fn)
        if return_dtype is None:
            if "return" in type_hints:
                self.return_dtype = DataType._infer_type(type_hints["return"])
            else:
                raise ValueError("return_dtype is required when function has no return annotation")
        else:
            self.return_dtype = return_dtype

    @overload
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T: ...
    @overload
    def __call__(self, *args: Any, **kwargs: Any) -> Expression: ...

    def __call__(self, *args: Any, **kwargs: Any) -> Expression | T:
        expr_args = []
        for arg in args:
            if isinstance(arg, Expression):
                expr_args.append(arg)
        for arg in kwargs.values():
            if isinstance(arg, Expression):
                expr_args.append(arg)

        if len(expr_args) == 0:
            # all inputs are Python literals, evaluate immediately
            return self._inner(*args, **kwargs)
        else:
            return Expression._row_wise_udf(self.name, self._inner, self.return_dtype, (args, kwargs), expr_args)


def call_func_with_evaluated_exprs(
    fn: Callable[..., Any],
    return_dtype: PyDataType,
    original_args: tuple[tuple[Any, ...], dict[str, Any]],
    evaluted_args: list[Any],
) -> PySeries:
    args, kwargs = original_args

    new_args = [evaluted_args.pop(0) if isinstance(arg, Expression) else arg for arg in args]
    new_kwargs = {key: (evaluted_args.pop(0) if isinstance(arg, Expression) else arg) for key, arg in kwargs.items()}

    output = fn(*new_args, **new_kwargs)
    dtype = DataType._from_pydatatype(return_dtype)
    return Series.from_pylist([output], dtype=dtype)._series



from __future__ import annotations

import pytest

import daft


def write_test_files(tmpdir, num_files: int) -> None:
    for i in range(num_files):
        file_data = {
            "id": list(range(i * 100, (i + 1) * 100)),
        }
        df = daft.from_pydict(file_data)
        df.write_parquet(f"{tmpdir}/file_{i}.parquet")


def test_sharding_should_error_for_non_file_scans() -> None:
    """Test that sharding should error for non-file scans."""
    df = daft.from_pydict({"id": list(range(100)), "value": [f"val_{i}" for i in range(100)]})
    with pytest.raises(ValueError, match="Sharding is not supported for in-memory sources"):
        df._shard(strategy="file", world_size=5, rank=0).collect()


def test_sharding_invalid_arguments() -> None:
    df = daft.from_pydict({"a": [1, 2, 3]})

    # Test invalid strategy.
    with pytest.raises(ValueError, match="Only file-based sharding is supported"):
        df._shard(strategy="invalid", world_size=5, rank=0)

    # Test invalid world_size.
    with pytest.raises(ValueError, match="World size for sharding must be greater than zero"):
        df._shard(strategy="file", world_size=0, rank=0)

    # Test invalid rank.
    with pytest.raises(ValueError, match="Rank must be less than the world size for sharding"):
        df._shard(strategy="file", world_size=5, rank=5)


def test_sharding_with_file_scan(tmpdir) -> None:
    num_files = 100
    write_test_files(tmpdir, num_files)

    world_size = 10
    all_files = set()

    for rank in range(world_size):
        sharded_df = daft.read_parquet(f"{tmpdir}/**/*.parquet", file_path_column="file_path")._shard(
            strategy="file", world_size=world_size, rank=rank
        )
        results = sharded_df.select("file_path").distinct().collect()
        assert len(results) > 0, "Shards should be somewhat fairly distributed and have at least one file"
        for result in results:
            file_path = result["file_path"]
            assert file_path not in all_files, f"File {file_path} should only be assigned to one shard"
            all_files.add(file_path)

    assert len(all_files) == num_files, "All files should be assigned to some shard"


def test_sharding_distribution_fairness(tmpdir) -> None:
    num_files = 100
    write_test_files(tmpdir, num_files)

    world_size = 10
    files_per_shard = []

    for rank in range(world_size):
        sharded_df = daft.read_parquet(f"{tmpdir}/**/*.parquet", file_path_column="file_path")._shard(
            strategy="file", world_size=world_size, rank=rank
        )
        files_per_shard.append(sharded_df.select("file_path").distinct().count_rows())

    avg_files = sum(files_per_shard) / len(files_per_shard)

    # The distribution should be reasonably uniform.
    variance = sum((x - avg_files) ** 2 for x in files_per_shard) / len(files_per_shard)
    std_dev = variance**0.5
    coefficient_of_variation = std_dev / avg_files if avg_files > 0 else float("inf")
    assert (
        coefficient_of_variation <= 0.5
    ), f"The distribution for sharding is too variable with a coefficient of variation of {coefficient_of_variation:.2f}"


def test_sharding_consistency(tmpdir) -> None:
    num_files = 50
    for i in range(num_files):
        file_data = {
            "row_id": list(range(i * 100, (i + 1) * 100)),
        }
        df = daft.from_pydict(file_data)
        df.write_parquet(f"{tmpdir}/file_{i}.parquet")

    world_size = 10

    for rank in range(world_size):
        canonical_result = (
            daft.read_parquet(f"{tmpdir}/**/*.parquet", file_path_column="file_path")
            ._shard(strategy="file", world_size=world_size, rank=rank)
            .to_arrow()
        )
        # Shard contents and order should be consistent across runs.
        for _ in range(3):
            new_result = (
                daft.read_parquet(f"{tmpdir}/**/*.parquet", file_path_column="file_path")
                ._shard(strategy="file", world_size=world_size, rank=rank)
                .to_arrow()
            )
            assert canonical_result == new_result, "Sharding result is inconsistent between runs"


def test_sharding_with_pushdowns(tmpdir) -> None:
    num_files = 100
    write_test_files(tmpdir, num_files)

    sharded_df = (
        daft.read_parquet(f"{tmpdir}/**/*.parquet", file_path_column="file_path")
        .limit(10)
        .sort("id")
        .distinct()
        ._shard(strategy="file", world_size=3, rank=0)
    )

    assert sharded_df.count_rows() == 10



from __future__ import annotations

from daft.udf.row_wise import RowWiseUdf
from typing import overload, Callable, TypeVar, TYPE_CHECKING
import sys

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

import functools
from daft.udf.legacy import udf, UDF

P = ParamSpec("P")
T = TypeVar("T")

if TYPE_CHECKING:
    from daft.datatype import DataType


class _DaftFuncDecorator:
    """`@daft.func` Decorator to convert a Python function into a `UDF`.

    Unlike `@daft.udf(...)`, `@daft.func` operates on single values instead of batches.

    Examples:
    >>> import daft
    >>> from daft import col
    >>> @daft.func  # or @daft.func()
    ... # or you can specify the return type like @daft.udf
    ... # @daft.func(return_dtype=daft.DataType.int64())
    ... def my_sum(a: int, b: int) -> int:
    ...     return a + b
    >>>
    >>> df = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
    >>> df.select(my_sum(col("x"), col("y"))).collect()
    ╭───────╮
    │ x     │
    │ ---   │
    │ Int64 │
    ╞═══════╡
    │ 5     │
    ├╌╌╌╌╌╌╌┤
    │ 7     │
    ├╌╌╌╌╌╌╌┤
    │ 9     │
    ╰───────╯
    <BLANKLINE>
    (Showing first 3 of 3 rows)
    """

    @overload
    def __new__(cls, *, return_dtype: DataType | None = None) -> Callable[[Callable[P, T]], RowWiseUdf[P, T]]: ...  # type: ignore
    @overload
    def __new__(cls, fn: Callable[P, T], *, return_dtype: DataType | None = None) -> RowWiseUdf[P, T]: ...  # type: ignore

    def __new__(  # type: ignore
        cls, fn: Callable[P, T] | None = None, *, return_dtype: DataType | None = None
    ) -> RowWiseUdf[P, T] | Callable[[Callable[P, T]], RowWiseUdf[P, T]]:
        def partial_udf(fn: Callable[P, T]) -> RowWiseUdf[P, T]:
            return RowWiseUdf(fn, return_dtype=return_dtype)

        return partial_udf if fn is None else partial_udf(fn)


__all__ = ["UDF", "udf"]



from __future__ import annotations

import numpy as np
import pyarrow as pa
import pytest

from daft import DataType, col
from daft.logical.schema import Schema
from daft.recordbatch import MicroPartition


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict({"a": pa.array([], type=pa.int64())}),  # 1 empty table
        MicroPartition.empty(Schema.from_pyarrow_schema(pa.schema({"a": pa.int64()}))),  # No tables
    ],
)
def test_micropartitions_size_bytes_empty(mp) -> None:
    assert mp.size_bytes() == 0


@pytest.mark.parametrize(
    "mp",
    [
        MicroPartition.from_pydict({"a": [1, 3, 2, 4]}),  # 1 table
        MicroPartition.concat(
            [
                MicroPartition.from_pydict({"a": np.array([]).astype(np.int64)}),
                MicroPartition.from_pydict({"a": [1]}),
                MicroPartition.from_pydict({"a": [3, 2, 4]}),
            ]
        ),  # 3 tables
    ],
)
def test_micropartitions_size_bytes(mp) -> None:
    assert mp.size_bytes() == (4 * 8)


def test_table_size_bytes() -> None:
    data = MicroPartition.from_pydict(
        {"a": [1, 2, 3, 4, None], "b": [False, True, False, True, None]}
    ).eval_expression_list([col("a").cast(DataType.int64()), col("b")])
    assert data.size_bytes() == (5 * 8 + 1) + (1 + 1)


@pytest.mark.parametrize("length", [10, 200])
def test_table_size_bytes_pyobj(length) -> None:
    import pickle

    size_per_obj = len(pickle.dumps(object()))
    data = MicroPartition.from_pydict({"a": [object()] * length})
    assert data.size_bytes() == size_per_obj * length



from __future__ import annotations

import base64
import io
from typing import Any, Callable, TypeVar

from daft.dependencies import np, pil_image

HookClass = TypeVar("HookClass")

_VIZ_HOOKS_REGISTRY = {}
_NUMPY_REGISTERED = False
_PILLOW_REGISTERED = False


def register_viz_hook(klass: type[HookClass], hook: Callable[[Any], str]) -> None:
    """Registers a visualization hook that returns the appropriate HTML for visualizing a specific class in HTML."""
    _VIZ_HOOKS_REGISTRY[klass] = hook


def get_viz_hook(val: object) -> Callable[[object], str] | None:
    global _NUMPY_REGISTERED
    global _PILLOW_REGISTERED
    if np.module_available() and not _NUMPY_REGISTERED:  # type: ignore[attr-defined]

        def _viz_numpy(val: np.ndarray[Any, Any]) -> str:
            return f"&ltnp.ndarray<br>shape={val.shape}<br>dtype={val.dtype}&gt"

        register_viz_hook(np.ndarray, _viz_numpy)
        _NUMPY_REGISTERED = True

    if pil_image.module_available() and not _PILLOW_REGISTERED:

        def _viz_pil_image(val: pil_image.Image) -> str:
            img = val.copy()
            img.thumbnail((128, 128))
            bio = io.BytesIO()
            img.save(bio, "JPEG")
            base64_img = base64.b64encode(bio.getvalue())
            return f'<img style="max-height:128px;width:auto" src="data:image/png;base64, {base64_img.decode("utf-8")}" alt="{val!s}" />'

        register_viz_hook(pil_image.Image, _viz_pil_image)
        _PILLOW_REGISTERED = True

    for klass in _VIZ_HOOKS_REGISTRY:
        if isinstance(val, klass):
            return _VIZ_HOOKS_REGISTRY[klass]
    return None
